import os
from argparse import ArgumentParser, RawTextHelpFormatter

import sbws.core.cleanup
import sbws.core.flowctrl2
import sbws.core.generate
import sbws.core.scanner
import sbws.core.stats
from sbws import __version__


def _default_dot_sbws_dname():
    home = os.path.expanduser("~")
    return os.path.join(home, ".sbws")


def create_parser():
    p = ArgumentParser(formatter_class=RawTextHelpFormatter)
    p.add_argument(
        "--version",
        action="version",
        help="sbws version",
        version="{}".format(__version__),
    )
    p.add_argument(
        "--log-level",
        choices=["debug", "info", "warning", "error", "critical"],
        help="Override the sbws log level",
    )
    p.add_argument("-c", "--config", help="Path to the sbws config file")
    sub = p.add_subparsers(dest="command")
    sbws.core.cleanup.gen_parser(sub)
    sbws.core.scanner.gen_parser(sub)
    sbws.core.generate.gen_parser(sub)
    sbws.core.stats.gen_parser(sub)
    sbws.core.flowctrl2.gen_parser(sub)
    return p



# -*- coding: utf-8 -*-
"""Test timestamp conversion util functions"""
from datetime import datetime, timedelta, timezone

from sbws.util.timestamp import (
    dt_obj_to_isodt_str,
    is_old,
    unixts_to_dt_obj,
    unixts_to_isodt_str,
    unixts_to_str,
)

isodt_str = "2018-05-23T12:55:04"
dt_obj = datetime.strptime(isodt_str, "%Y-%m-%dT%H:%M:%S")
unixts = int(dt_obj.replace(tzinfo=timezone.utc).timestamp())


def test_dt_obj_to_isodt_str():
    assert isodt_str == dt_obj_to_isodt_str(dt_obj)


def test_unixts_to_dt_obj():
    assert dt_obj == unixts_to_dt_obj(unixts)


def test_unixts_to_isodt_str():
    assert isodt_str == unixts_to_isodt_str(unixts)


def test_unixts_to_str():
    assert str(unixts) == unixts_to_str(unixts)


def test_is_old():
    # Since this timestamp is generated a few microseconds before checking
    # the oldest timestamp, it will be old.
    old_timestamp = datetime.utcnow() - timedelta(days=5)
    assert is_old(old_timestamp)
    # A recent timestamp should be at least 1 second newer that the oldest
    recent_timestamp = (
        datetime.utcnow() - timedelta(days=5) + timedelta(seconds=1)
    )
    assert not is_old(recent_timestamp)



import copy
import logging
import time
from decimal import Decimal

from ..globals import MAX_RECENT_PRIORITY_RELAY_COUNT
from ..lib.relaylist import RelayList
from ..lib.resultdump import ResultDump, ResultError
from ..util import state, timestamps

log = logging.getLogger(__name__)


class RelayPrioritizer:
    def __init__(self, args, conf, relay_list, result_dump):
        assert isinstance(relay_list, RelayList)
        assert isinstance(result_dump, ResultDump)
        self.fresh_seconds = (
            conf.getint("general", "data_period") * 24 * 60 * 60
        )
        self.relay_list = relay_list
        self.result_dump = result_dump
        self.measure_authorities = conf.getboolean(
            "relayprioritizer", "measure_authorities"
        )
        self.min_to_return = conf.getint("relayprioritizer", "min_relays")
        self.fraction_to_return = conf.getfloat(
            "relayprioritizer", "fraction_relays"
        )
        self._state = state.State(conf.getpath("paths", "state_fname"))
        self._recent_priority_list = timestamps.DateTimeSeq(
            [], 120, self._state, "recent_priority_list"
        )
        self._recent_priority_relay = timestamps.DateTimeIntSeq(
            [],
            MAX_RECENT_PRIORITY_RELAY_COUNT,
            self._state,
            "recent_priority_relay",
        )

    def increment_recent_priority_list(self):
        """
        Increment the number of times that
        :meth:`~sbws.lib.relayprioritizer.RelayPrioritizer.best_priority`
        has been run.
        """
        # NOTE: blocking, writes to file!
        self._recent_priority_list.update()

    @property
    def recent_priority_list_count(self):
        return len(self._recent_priority_list)

    def increment_recent_priority_relay(self, relays_count):
        """
        Increment the number of relays that have been "prioritized" to be
        measured in a
        :meth:`~sbws.lib.relayprioritizer.RelayPrioritizer.best_priority`.
        """
        # NOTE: blocking, writes to file!
        self._recent_priority_relay.update(number=relays_count)

    @property
    def recent_priority_relay_count(self):
        return len(self._recent_priority_relay)

    def best_priority(
        self, prioritize_result_error=False, return_fraction=True
    ):
        """Yields a new ordered list of relays to be measured next.

        The relays that were measured farther away in the past,
        get prioritized (lowest priority number, first in the list).
        The relays that were measured more recently get lower priority (last in
        the list, higher priority number).

        Optionally, the relays which measurements failed can be prioritized
        (first in the list).
        However, unstable relays that fail often to be measured, might fail
        again and stable relays will get measured only when their measurements
        become old enough.
        The opposite might be more suitable: give lower priority to the relays
        that are unstable, to don't spend time measuring relays that might fail
        to be measured.

        Optionally, return only a fraction of all the relays in the network.
        Since there could be new relays in the network while measuring the
        list of relays returned by this method, this method is run again
        before all the relays in the network are measured.

        .. note::

            In a future refactor, instead of having a static fraction of relays
            to be measured, this method could be call when it's known that
            there're X number of new relays in the network.

        Since measurements made before than X days ago (too old) are not
        considered, and the initial list of past measurements is only filtered
        when the scanner starts, it's needed to filter here again to discard
        those measurements.

        :param bool prioritize_result_error: whether prioritize or not
            measurements that did not succeed.
        :param bool return_fraction: whether to return only a fraction of the
            relays seen in the network or return all.

        return: a generator of the new ordered list of relays to measure next.

        """
        fn_tstart = Decimal(time.time())
        relays = set(copy.deepcopy(self.relay_list.relays))
        if not self.measure_authorities:
            relays = relays.difference(set(self.relay_list.authorities))
        # Since there will be new measurements every time this method is called
        # again, update the list of results.
        # In a future refactor with other data structure there should not be
        # needed.
        rd = self.result_dump
        for relay in relays:
            results = rd.results_for_relay(relay)
            priority = 0
            # The time before which we do not consider results valid anymore
            oldest_allowed = time.time() - self.fresh_seconds
            for result in results:
                # Ignore results that are too far in the past
                if result.time < oldest_allowed:
                    continue
                # Calculate freshness as the remaining time until this result
                # is no longer valid
                freshness = result.time - oldest_allowed
                if (
                    isinstance(result, ResultError)
                    and prioritize_result_error is True
                ):
                    # log.debug('Cutting freshness for a %s result by %d%% for'
                    #           ' %s', result.type.value,
                    #           result.freshness_reduction_factor * 100,
                    #           relay.nickname)
                    # result.freshness_reduction_factor are hard-coded values
                    # on how much prioritize measurements that failed
                    # depending on the type of error.
                    # In a future refactor, create these values on an algorithm
                    # or create constants.
                    freshness *= max(
                        1.0 - result.freshness_reduction_factor, 0
                    )
                priority += freshness
            # In a future refactor, do not create a new attribute
            relay.priority = priority
        # Sort the relays by their priority, with the smallest (best) priority
        # relays at the front
        relays = sorted(relays, key=lambda r: r.priority)

        fn_tstop = Decimal(time.time())
        fn_tdelta = (fn_tstop - fn_tstart) * 1000
        log.info("Spent %f msecs calculating relay best priority", fn_tdelta)

        # Return a fraction of relays in the network if return_fraction is
        # True, otherwise return all.
        cutoff = max(
            int(len(relays) * self.fraction_to_return), self.min_to_return
        )
        upper_limit = cutoff if return_fraction else len(relays)
        # NOTE: these two are blocking, write to disk
        # Increment the number of times ``best_priority`` has been run.
        self.increment_recent_priority_list()
        # Increment the number of relays that have been "prioritized".
        # Because in a small testing network the upper limit could be smaller
        # than the number of relays in the network, use the length of the list.
        self.increment_recent_priority_relay(len(relays[0:upper_limit]))
        for relay in relays[0:upper_limit]:
            log.debug(
                "Returning next relay %s with priority %f",
                relay.nickname,
                relay.priority,
            )
            # In a future refactor, a new attribute should not be created,
            # then no need to remove it.
            del relay.priority
            # Increment the number of times a really was "prioritized" to be
            # measured.
            relay.increment_relay_recent_priority_list()
            yield relay



#!/usr/bin/env python3
""""""
import argparse

from sbws.lib.bwfile_health import BwFile


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--file-path", help="Bandwidth file path.")

    args = parser.parse_args()

    header_health = BwFile.load(args.file_path)
    header_health.report


if __name__ == "__main__":
    main()



"""Utils file system functions"""

import logging
import shutil

log = logging.getLogger(__name__)

DISK_SPACE_TEXT = """
Disk space requirements
-----------------------
v3bw files: the maximum space required is ~{mb_bw} MB, after {d_bw} days.
result files: the maximum space required is ~{mb_results} MB, after {d_r} days.
tor directory: the space required is ~{mb_tor} MB.
code and dependencies: the space required is ~{mb_code} MB
Total disk space required is: ~{mb_total} MB
"""


def sbws_required_disk_space(conf):
    """Disk space required by sbws files.
    Rough calculations.

    :param ConfigParser conf: sbws configuration
    :returns: int, size in MB
    """
    text_dict = {}
    # Number of relays per line average size in Bytes
    size_v3bw_file = 7500 * 220
    # default crontab configuration will run genenerate every hour
    num_v3bw_files_day = 24
    # ~1000 is the length of a line when the result is successful
    # ~4550 is the number of lines of the biggest result file
    size_result_file = 4550 * 1000
    num_result_files_day = 1
    space_v3bw_files_day = size_v3bw_file * num_v3bw_files_day
    space_result_files_day = size_result_file * num_result_files_day
    size_compressed_files = 600 * 1024
    # default crontab configuration will run cleanup once a day
    # default cleanup configuration will compress v3bw files after 1 day
    # and delete them after 7 days
    v3bw_compress_after_days = conf.getint(
        "cleanup", "v3bw_files_compress_after_days"
    )
    v3bw_delete_after_days = conf.getint(
        "cleanup", "v3bw_files_delete_after_days"
    )
    v3bw_max_space_after_delete = (
        space_v3bw_files_day * v3bw_compress_after_days
    ) + (size_compressed_files * num_v3bw_files_day * v3bw_delete_after_days)
    text_dict["mb_bw"] = round(v3bw_max_space_after_delete / 1000**2)
    text_dict["d_bw"] = v3bw_delete_after_days
    # default crontab configuration will run cleanup once a day
    # default cleanup configuration will compress v3bw files after 1 day
    # and delete them after 7 days
    results_compress_after_days = conf.getint(
        "cleanup", "data_files_compress_after_days"
    )
    results_delete_after_days = conf.getint(
        "cleanup", "data_files_delete_after_days"
    )
    results_max_space_after_delete = (
        space_result_files_day * results_compress_after_days
    ) + (
        size_compressed_files * num_v3bw_files_day * results_delete_after_days
    )
    text_dict["mb_results"] = round(results_max_space_after_delete / 1000**2)
    text_dict["d_r"] = results_delete_after_days
    # not counted rotated files and assuming that when it is not rotated the
    # size will be approximately 10MiB
    space_log_files = 0
    if conf.getboolean("logging", "to_file"):
        size_log_file = conf.getint("logging", "to_file_max_bytes")
        num_log_files = conf.getint("logging", "to_file_num_backups")
        space_log_files = size_log_file * num_log_files
    text_dict["mb_log"] = space_log_files
    # roughly, size of a current tor dir
    size_tor_dir = 19828000
    text_dict["mb_tor"] = round(size_tor_dir / 1000**2)
    # roughly, the size of this code and dependencies
    size_code_deps = 2097152
    text_dict["mb_code"] = round(size_code_deps / 1000**2)
    # Multiply per 2, just in case
    size_total = (
        results_max_space_after_delete
        + v3bw_max_space_after_delete
        + space_log_files
        + size_tor_dir
        + size_code_deps
    ) * 2
    text_dict["mb_total"] = round(size_total / 1000**2)
    space_text = DISK_SPACE_TEXT.format(**text_dict)
    return space_text


def df(path):
    # Not being used, since it makes a disk space system call and some
    # systems might not allow it
    """Return space left on device where path is in MiB."""
    return round(shutil.disk_usage(path).free / (1024**2))


def is_low_space(conf):
    # Not being used, since it makes a disk space system call and some
    # systems might not allow it
    """Warn and return True when the space left on the device is less than
    what is needed for sbws and False otherwise needs.
    """
    disk_required_mb = sbws_required_disk_space(conf)
    disk_avail_mb = df(conf.getpath("paths", "sbws_home"))
    if disk_avail_mb < disk_required_mb:
        log.warn(
            "The space left on the device (%s MiB) is less than "
            "the minimum recommended to run sbws (%s MiB)."
            "Run sbws cleanup to delete old sbws generated files.",
            disk_avail_mb,
            disk_required_mb,
        )
        return True
    return False



"""Test that the KeyValues in a bandwidth file make sense."""
import os.path

from sbws.lib.bwfile_health import BwFile


def test_bwfile_health(root_data_path):
    bwfile = BwFile.load(
        os.path.join(root_data_path, "2020-03-22-08-35-00-bandwidth")
    )
    assert bwfile.header.is_correct
    assert bwfile.are_bwlines_correct
    assert bwfile.is_correct


def test_bwlines_health(capsys, root_data_path):
    bwfile = BwFile.load(
        os.path.join(root_data_path, "2020-03-22-08-35-00-bandwidth")
    )
    out = (
        "\nrelay_recent_measurement_attempt_count <= relay_recent_priority_list_count,\n"  # noqa
        "True\n"
        "relay_recent_priority_list_count <= relay_recent_consensus_count,\n"
        "True\n\n"
    )
    for bwline in bwfile.bwlines:
        bwline.report
        assert out == capsys.readouterr().out



"""Expected bandwidth file values for KeyValues."""
# flake8: noqa: E741
# (E741 ambiguous variable name), when using l.
import logging

from stem import descriptor

from sbws.globals import MAX_RECENT_CONSENSUS_COUNT  # 120
from sbws.globals import MAX_RECENT_PRIORITY_LIST_COUNT  # 120
from sbws.globals import MAX_RECENT_PRIORITY_RELAY_COUNT  # 48000
from sbws.globals import FRACTION_RELAYS, PERIOD_DAYS
from sbws.lib.v3bwfile import BWLINE_KEYS_V1_4, HEADER_INT_KEYS

logging.basicConfig(
    level=logging.INFO,
)
logger = logging.getLogger(__name__)


# Based on observation
MAX_HOURS_PRIORITY_LIST = 5
MIN_RECENT_CONSENSUS_COUNT = PERIOD_DAYS * 12  # 60
MIN_RELAYS = 6000
# 24
MIN_RECENT_PRIORITY_LIST_COUNT = PERIOD_DAYS * 24 / MAX_HOURS_PRIORITY_LIST
MIN_RELAYS_PER_PRIORITY_LIST = int(MIN_RELAYS * FRACTION_RELAYS)  # 300
# 7200
MIN_RECENT_PRIORITY_RELAY_COUNT = (
    MIN_RECENT_PRIORITY_LIST_COUNT * MIN_RELAYS_PER_PRIORITY_LIST
)
# If the number of attempts is not equal to the number of relays being in the
# priority list, there's a bug.
MIN_RECENT_MEASUREMENT_ATTEMPT_COUNT = MIN_RECENT_PRIORITY_RELAY_COUNT
MAX_RECENT_MEASUREMENT_ATTEMPT_COUNT = MAX_RECENT_PRIORITY_RELAY_COUNT

# noqa
REPORT_TEMPLATE_BWFILE = (
    "sum(relay_recent_measurement_attempt_count) "
    "<= recent_measurement_attempt_count, "
    "{self.is_sum_relay_recent_measurement_attempt_count_lte_recent_measurement_attempt_count}\n"  # noqa
)

REPORT_TEMPLATE_BWHEADER = """
Header,
recent_consensus_count >= min, {self.is_consensus_gte_min}
recent_consensus_count <= max, {self.is_consensus_lte_max}
recent_priority_list_count >= min, {self.is_priority_list_gte_min}
recent_priority_list_count <= max, {self.is_priority_list_lte_max}
recent_priority_relay_count >= min, {self.is_priority_relay_gte_min}
recent_priority_relay_count <= max, {self.is_priority_relay_lte_max}
""" + (
    "recent_measurement_attempt_count >= min, "
    "{self.is_measurement_attempt_gte_min}\n"
    "recent_measurement_attempt_count <= max, "
    "{self.is_measurement_attempt_lte_max}\n"
    "recent_measurement_attempt_count == recent_priority_relay_count, "
    "{self.is_attempt_e_priority_relay}\n"
    "recent_measurement_attempt_count >= total excluded, "
    "{self.is_attempt_gte_failure_exclude}\n"
)


REPORT_TEMPLATE_BWLINES = """
relays correct, {self.are_bwlines_correct}
"""

REPORT_TEMPLATE_BWLINE = """
relay_recent_measurement_attempt_count <= relay_recent_priority_list_count,
{self.is_relay_recent_measurement_attempt_count_lte_relay_recent_priority_list_count}
relay_recent_priority_list_count <= relay_recent_consensus_count,
{self.is_relay_recent_priority_list_count_lte_relay_recent_consensus_count}
"""


class BwFile:
    def __init__(self, header, bwlines):
        self.header = BwHeader(header)
        self.bwlines = [BwLine(line) for line in bwlines]

    @classmethod
    def load(cls, file_path):
        logger.info("Parsing content of %s.", file_path)
        document = descriptor.parse_file(file_path)
        bwfiles = list(document)
        if bwfiles:
            # When parsing one file, there is only 1 bwfile
            bwfile = bwfiles[0]
            return cls(bwfile.header, bwfile.measurements.values())

    @property
    def sum_relay_recent_measurement_attempt_count(self):
        return sum(
            [l.relay_recent_measurement_attempt_count for l in self.bwlines]
        )

    @property
    def is_sum_relay_recent_measurement_attempt_count_lte_recent_measurement_attempt_count(  # noqa
        self,
    ):
        return (
            self.sum_relay_recent_measurement_attempt_count
            <= self.header.recent_measurement_attempt_count
        )

    @property
    def are_bwlines_correct(self):
        return not list(filter(lambda x: not x.is_correct, self.bwlines))

    @property
    def is_correct(self):
        methods = [m for m in dir(self) if m.startswith("is_")]
        methods.remove("is_correct")
        return not list(filter(lambda x: not getattr(self, x), methods))

    @property
    def report(self):
        print(REPORT_TEMPLATE_BWFILE.format(self=self))
        self.header.report
        print(REPORT_TEMPLATE_BWLINES.format(self=self))


class BwLine:
    def __init__(self, line):
        for k, v in line.items():
            if k in BWLINE_KEYS_V1_4:
                setattr(self, k, int(v))
            else:
                setattr(self, k, v)

    @property
    def is_relay_recent_priority_list_count_lte_relay_recent_consensus_count(
        self,
    ):
        return (
            self.relay_recent_priority_list_count
            <= self.relay_in_recent_consensus_count
        )

    @property
    def is_relay_recent_measurement_attempt_count_lte_relay_recent_priority_list_count(  # noqa
        self,
    ):
        return (
            self.relay_recent_measurement_attempt_count
            <= self.relay_recent_priority_list_count
        )

    def is_relay_recent_consensus_count_lte_recent_consensus_count(
        self, recent_consensus_count
    ):
        return self.relay_in_recent_consensus_count <= recent_consensus_count

    @property
    def is_correct(self):
        methods = [m for m in dir(self) if m.startswith("is_")]
        methods.remove("is_correct")
        return not list(filter(lambda x: not getattr(self, x), methods))

    @property
    def report(self):
        print(REPORT_TEMPLATE_BWLINE.format(self=self))


class BwHeader:
    def __init__(self, header):
        for k, v in header.items():
            if k in HEADER_INT_KEYS:
                setattr(self, k, int(v))
            else:
                setattr(self, k, v)
        # logger.info(self.__dict__)

    @classmethod
    def load(cls, file_path):
        logger.info("Parsing content of %s.", file_path)
        document = descriptor.parse_file(file_path)
        bwfiles = list(document)
        if bwfiles:
            bwfile = bwfiles[0]
            return cls(bwfile.header)

    @property
    def is_consensus_lte_max(self):
        return self.recent_consensus_count <= MAX_RECENT_CONSENSUS_COUNT

    @property
    def is_consensus_gte_min(self):
        return self.recent_consensus_count >= MIN_RECENT_CONSENSUS_COUNT

    @property
    def is_priority_list_lte_max(self):
        return (
            self.recent_priority_list_count <= MAX_RECENT_PRIORITY_LIST_COUNT
        )

    @property
    def is_priority_list_gte_min(self):
        return (
            self.recent_priority_list_count >= MIN_RECENT_PRIORITY_LIST_COUNT
        )

    @property
    def is_priority_relay_lte_max(self):
        return (
            self.recent_priority_relay_count <= MAX_RECENT_PRIORITY_RELAY_COUNT
        )

    @property
    def is_priority_relay_gte_min(self):
        return (
            self.recent_priority_relay_count >= MIN_RECENT_PRIORITY_RELAY_COUNT
        )

    @property
    def is_measurement_attempt_gte_min(self):
        return (
            self.recent_measurement_attempt_count
            >= MIN_RECENT_MEASUREMENT_ATTEMPT_COUNT
        )

    @property
    def is_measurement_attempt_lte_max(self):
        return (
            self.recent_measurement_attempt_count
            <= MAX_RECENT_MEASUREMENT_ATTEMPT_COUNT
        )

    @property
    def is_attempt_e_priority_relay(self):
        return (
            self.recent_measurement_attempt_count
            == self.recent_priority_relay_count
        )

    @property
    def is_attempt_gte_failure(self):
        return (
            self.recent_measurement_attempt_count
            >= self.recent_measurement_failure_count
        )

    @property
    def total_excluded(self):
        return sum(
            [
                self.recent_measurements_excluded_error_count,
                self.recent_measurements_excluded_few_count,
                self.recent_measurements_excluded_near_count,
                self.recent_measurements_excluded_old_count,
            ]
        )

    @property
    def total_excluded_failure(self):
        return sum(
            [self.total_excluded, self.recent_measurement_failure_count]
        )

    @property
    def is_attempt_gte_failure_exclude(self):
        return (
            self.recent_measurement_attempt_count
            >= self.total_excluded_failure
        )

    @property
    def is_correct(self):
        methods = [m for m in dir(self) if m.startswith("is_")]
        methods.remove("is_correct")
        return not list(filter(lambda x: not getattr(self, x), methods))

    @property
    def report(self):
        print(REPORT_TEMPLATE_BWHEADER.format(self=self))



"""Common pytest configuration for unit and integration tests."""
import os.path
from unittest import mock

import pytest
from freezegun import freeze_time
from stem import descriptor

from sbws import settings
from sbws.lib import relaylist, relayprioritizer, resultdump
from sbws.util.parser import create_parser


@pytest.fixture(scope="session")
def parser():
    return create_parser()


@pytest.fixture()
def datadir(request):
    """get, read, open test files from the tests relative "data" directory."""

    class D:
        def __init__(self, basepath):
            self.basepath = basepath

        def open(self, name, mode="r"):
            return self.basepath.join(name).open(mode)

        def join(self, name):
            return self.basepath.join(name).strpath

        def read(self, name):
            with self.open(name, "r") as f:
                return f.read()

        def readlines(self, name):
            with self.open(name, "r") as f:
                return f.readlines()

    return D(request.fspath.dirpath("data"))


@pytest.fixture(scope="session")
def root_data_path():
    """Path to the data dir in the tests root, for both unit and integration
    tests.
    """
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data",
    )


@pytest.fixture(scope="session")
def router_statuses(root_data_path):
    p = os.path.join(root_data_path, "2020-02-29-10-00-00-consensus")
    network_statuses = descriptor.parse_file(p)
    network_statuses_list = list(network_statuses)
    return network_statuses_list


@pytest.fixture(scope="session")
def router_statuses_1h_later(root_data_path):
    p = os.path.join(root_data_path, "2020-02-29-11-00-00-consensus")
    network_statuses = descriptor.parse_file(p)
    network_statuses_list = list(network_statuses)
    return network_statuses_list


@pytest.fixture(scope="session")
def router_statuses_5days_later(root_data_path):
    p = os.path.join(root_data_path, "2020-03-05-10-00-00-consensus")
    network_statuses = descriptor.parse_file(p)
    network_statuses_list = list(network_statuses)
    return network_statuses_list


@pytest.fixture(scope="session")
def controller(router_statuses):
    controller = mock.Mock()
    controller.get_info.return_value = "params foo=23"
    controller.get_network_statuses.return_value = router_statuses
    return controller


@pytest.fixture(scope="session")
def controller_1h_later(router_statuses_1h_later):
    controller = mock.Mock()
    controller.get_network_statuses.return_value = router_statuses_1h_later
    return controller


@pytest.fixture(scope="session")
def controller_5days_later(router_statuses_5days_later):
    controller = mock.Mock()
    controller.get_network_statuses.return_value = router_statuses_5days_later
    return controller


@pytest.fixture(scope="session")
def server_descriptors(root_data_path):
    p = os.path.join(root_data_path, "2020-02-29-10-05-00-server-descriptors")
    server_descriptors = descriptor.parse_file(p)
    server_descriptors_list = list(server_descriptors)
    return server_descriptors_list


@pytest.fixture(scope="session")
def server_descriptor(server_descriptors):
    return server_descriptors[0]


@pytest.fixture(scope="session")
def router_status(server_descriptor, router_statuses):
    rs = [
        ns
        for ns in router_statuses
        if ns.fingerprint == server_descriptor.fingerprint
    ][0]
    return rs


# Because of the function scoped `args` in `tests.unit.conftest`, this has to
# be function scoped too.
@pytest.fixture(scope="function")
def relay_list(args, conf, controller):
    """Returns a RelayList containing the Relays in the controller"""
    with freeze_time("2020-02-29 10:00:00"):
        return relaylist.RelayList(args, conf, controller)


@pytest.fixture(scope="function")
def result_dump(args, conf):
    """Returns a ResultDump without Results"""
    # To stop the thread that would be waiting for new results
    settings.set_end_event()
    return resultdump.ResultDump(args, conf)


@pytest.fixture(scope="function")
def relay_prioritizer(args, conf_results, relay_list, result_dump):
    """
    Returns a RelayPrioritizer with a RelayList and a ResultDump.
    """
    return relayprioritizer.RelayPrioritizer(
        args, conf_results, relay_list, result_dump
    )



"""Unit tests for sbws.core.generate module."""
import argparse

from sbws.core.generate import gen_parser
from sbws.globals import PROP276_ROUND_DIG, TORFLOW_ROUND_DIG


def test_gen_parser_arg_round_digs():
    """
    Test that both --torflow-round-digs and --round-digs arguments can be
    passed and round-digs is PROP276_ROUND_DIG by default.

    """
    parent_parser = argparse.ArgumentParser(prog="sbws")
    subparsers = parent_parser.add_subparsers(help="generate help")
    parser_generate = gen_parser(subparsers)
    # Explicitly set empty arguments, otherwise pytest will use pytest
    # arguments
    args = parser_generate.parse_args([])
    assert args.round_digs == PROP276_ROUND_DIG
    # torflow_round_digs is not in the Namespace
    assert getattr(args, "torflow_round_digs", None) is None
    # but it can still be passed as an argument
    args = parser_generate.parse_args(
        ["--torflow-round-digs", str(TORFLOW_ROUND_DIG)]
    )
    # though the variable is named round_digs
    assert args.round_digs == TORFLOW_ROUND_DIG
    # or use the short version
    args = parser_generate.parse_args(["-r", str(TORFLOW_ROUND_DIG)])
    assert args.round_digs == TORFLOW_ROUND_DIG
    # or use round-digs explicitly
    args = parser_generate.parse_args(["--round-digs", str(PROP276_ROUND_DIG)])
    assert args.round_digs == PROP276_ROUND_DIG



"""Unit tests for stem.py"""

from sbws.util.stem import parse_user_torrc_config


def test_parse_user_torrc_config_new_keyvalue_options_success():
    config_torrc_extra_lines = """
    Log debug file /tmp/tor-debug.log
    NumCPUs 1
    """
    torrc_dict = parse_user_torrc_config({}, config_torrc_extra_lines)
    assert torrc_dict == {
        "Log": "debug file /tmp/tor-debug.log",
        "NumCPUs": "1",
    }


def test_parse_user_torrc_config_existing_keyvalue_options_fail(caplog):
    torrc_dict = {"SocksPort": "auto"}
    config_torrc_extra_lines = """
    SocksPort 9050
    """
    torrc_dict_new = parse_user_torrc_config(
        torrc_dict, config_torrc_extra_lines
    )
    # the new dictionary contains the existing key option and a list with both
    # the existing value and the new value
    assert torrc_dict_new != torrc_dict
    assert torrc_dict_new == {"SocksPort": ["auto", "9050"]}


def test_parse_user_torrc_config_new_key_option_success():
    config_torrc_extra_lines = """
    LongLivedPorts
    """
    torrc_dict = parse_user_torrc_config({}, config_torrc_extra_lines)
    assert torrc_dict == {"LongLivedPorts": None}



from statistics import mean


def bw_measurements_from_results(results):
    return [
        dl["amount"] / dl["duration"] for r in results for dl in r.downloads
    ]


def bw_filt(bw_measurements):
    """Filtered bandwidth for a relay.

    It is the equivalent to Torflow's ``filt_sbw``.
    ``mu`` in this function is the equivalent to Torflow's ``sbw``.
    """
    # It's safe to return 0 here, because:
    # 1. this value will be the numerator when calculating the ratio.
    # 2. `kb_round_x_sig_dig` returns a minimum of 1.
    # This should never be the case, as the measurements come from successful
    # results.
    if not bw_measurements:
        return 0
    # Torflow is rounding to an integer, so is `bw_mean_from_results` in
    # `v3bwfile.py`
    mu = round(mean(bw_measurements))
    bws_gte_mean = list(filter(lambda bw: bw >= mu, bw_measurements))
    if bws_gte_mean:
        return round(mean(bws_gte_mean))
    return mu



# -*- coding: utf-8 -*-
"""Test generation of bandwidth measurements document (v3bw)"""
import json
import logging
import math
import os.path
from unittest import mock

import pytest

from sbws import __version__ as version
from sbws.globals import (
    MIN_REPORT,
    PROP276_ROUND_DIG,
    SBWS_SCALING,
    SPEC_VERSION,
    TORFLOW_ROUND_DIG,
    TORFLOW_SCALING,
)
from sbws.lib.resultdump import Result, ResultSuccess, load_result_file
from sbws.lib.v3bwfile import (
    HEADER_RECENT_MEASUREMENTS_EXCLUDED_KEYS,
    KEYVALUE_SEP_V1,
    LINE_SEP,
    TERMINATOR,
    V3BWFile,
    V3BWHeader,
    V3BWLine,
    num_results_of_type,
    round_sig_dig,
)
from sbws.util.state import CustomDecoder
from sbws.util.timestamp import now_fname, now_isodt_str, now_unixts

timestamp = 1523974147
timestamp_l = str(timestamp)
version_l = KEYVALUE_SEP_V1.join(["version", SPEC_VERSION])
scanner_country = "US"
scanner_country_l = KEYVALUE_SEP_V1.join(["scanner_country", scanner_country])
destinations_countries = "00,DE"
destinations_countries_l = KEYVALUE_SEP_V1.join(
    ["destinations_countries", destinations_countries]
)
software_l = KEYVALUE_SEP_V1.join(["software", "sbws"])
software_version_l = KEYVALUE_SEP_V1.join(["software_version", version])
file_created = "2018-04-25T13:10:57"
file_created_l = KEYVALUE_SEP_V1.join(["file_created", file_created])
latest_bandwidth = "2018-04-17T14:09:07"
latest_bandwidth_l = KEYVALUE_SEP_V1.join(
    ["latest_bandwidth", latest_bandwidth]
)
attempts = "1"
attempts_l = KEYVALUE_SEP_V1.join(
    ["recent_measurement_attempt_count", attempts]
)
failure = "0"
failure_l = KEYVALUE_SEP_V1.join(["recent_measurement_failure_count", failure])
header_ls = [
    timestamp_l,
    version_l,
    destinations_countries_l,
    file_created_l,
    latest_bandwidth_l,
    # attempts_l, failure_l,
    scanner_country_l,
    software_l,
    software_version_l,
    TERMINATOR,
]
header_str = LINE_SEP.join(header_ls) + LINE_SEP
earliest_bandwidth = "2018-04-16T14:09:07"
earliest_bandwidth_l = KEYVALUE_SEP_V1.join(
    ["earliest_bandwidth", earliest_bandwidth]
)
generator_started = "2018-04-16T14:09:05"
generator_started_l = KEYVALUE_SEP_V1.join(
    ["generator_started", generator_started]
)
tor_version = "0.4.2.5"
tor_version_l = KEYVALUE_SEP_V1.join(["tor_version", tor_version])

header_extra_ls = [
    timestamp_l,
    version_l,
    earliest_bandwidth_l,
    file_created_l,
    generator_started_l,
    latest_bandwidth_l,
    software_l,
    software_version_l,
    tor_version_l,
    TERMINATOR,
]
header_extra_str = LINE_SEP.join(header_extra_ls) + LINE_SEP

# Line produced without any scaling.
# unmeasured and vote are not congruent with the exclusion,
# but `from_data` is only used in the test and doesn't include the
# arg `min_num`
raw_bwl_str = (
    "bw=56 bw_mean=61423 bw_median=55656 "
    "consensus_bandwidth=600000 consensus_bandwidth_is_unmeasured=False "
    "desc_bw_avg=1000000000 desc_bw_bur=123456 desc_bw_obs_last=524288 "
    "desc_bw_obs_mean=524288 error_circ=0 error_destination=0 error_misc=0 "
    "error_second_relay=0 error_stream=2 "
    "master_key_ed25519=g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s "
    "nick=A "
    "node_id=$AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA "
    "relay_in_recent_consensus_count=3 "
    "relay_recent_measurement_attempt_count=2 "
    "relay_recent_measurements_excluded_error_count=2 "
    "relay_recent_priority_list_count=3 "
    "rtt=456 success=1 "
    "time=2018-04-17T14:09:07\n"
)

v3bw_str = header_extra_str + raw_bwl_str


def test_v3bwheader_str():
    """Test header str"""
    header = V3BWHeader(
        timestamp_l,
        scanner_country=scanner_country,
        destinations_countries=destinations_countries,
        file_created=file_created,
    )
    assert header_str == str(header)


def test_v3bwheader_extra_str():
    """Test header str with additional headers"""
    header = V3BWHeader(
        timestamp_l,
        file_created=file_created,
        generator_started=generator_started,
        earliest_bandwidth=earliest_bandwidth,
        tor_version=tor_version,
    )
    assert header_extra_str == str(header)


def test_v3bwheader_from_lines():
    header_obj = V3BWHeader(
        timestamp_l,
        file_created=file_created,
        generator_started=generator_started,
        earliest_bandwidth=earliest_bandwidth,
        tor_version=tor_version,
    )
    header, _ = V3BWHeader.from_lines_v1(header_extra_ls)
    assert str(header_obj) == str(header)


def test_v3bwheader_from_text():
    header_obj = V3BWHeader(
        timestamp_l,
        file_created=file_created,
        generator_started=generator_started,
        earliest_bandwidth=earliest_bandwidth,
        tor_version=tor_version,
    )
    header, _ = V3BWHeader.from_text_v1(header_extra_str)
    assert str(header_obj) == str(header)


def test_num_results_of_type(result_success, result_error_stream):
    assert num_results_of_type([result_success], "success") == 1
    assert num_results_of_type([result_error_stream], "success") == 0
    assert num_results_of_type([result_success], "error-stream") == 0
    assert num_results_of_type([result_error_stream], "error-stream") == 1


def assert_round_sig_dig_any_digits(n, result):
    """Test that rounding n to any reasonable number of significant digits
    produces result."""
    max_digits_int64 = int(math.ceil(math.log10(2**64 - 1))) + 1
    for d in range(1, max_digits_int64 + 1):
        assert round_sig_dig(n, digits=d) == result


def assert_round_sig_dig_any_digits_error(n, elp_fraction=0.5):
    """Test that rounding n to any reasonable number of significant digits
    produces a result within elp_fraction * 10.0 ** -(digits - 1)."""
    max_digits_int64 = int(math.ceil(math.log10(2**64 - 1))) + 1
    for d in range(1, max_digits_int64 + 1):
        error_fraction = elp_fraction * (10.0 ** -(d - 1))
        # use ceil rather than round, to work around floating-point inaccuracy
        e = int(math.ceil(n * error_fraction))
        assert round_sig_dig(n, digits=d) >= n - e
        assert round_sig_dig(n, digits=d) <= n + e


def test_round_sig_dig():
    """Test rounding to a number of significant digits."""
    # Expected values
    assert round_sig_dig(11, 1) == 10
    assert round_sig_dig(11, 2) == 11

    assert round_sig_dig(15, 1) == 20
    assert round_sig_dig(15, 2) == 15

    assert round_sig_dig(54, 1) == 50
    assert round_sig_dig(54, 2) == 54

    assert round_sig_dig(96, 1) == 100
    assert round_sig_dig(96, 2) == 96

    assert round_sig_dig(839, 1) == 800
    assert round_sig_dig(839, 2) == 840
    assert round_sig_dig(839, 3) == 839

    assert round_sig_dig(5789, 1) == 6000
    assert round_sig_dig(5789, 2) == 5800
    assert round_sig_dig(5789, 3) == 5790
    assert round_sig_dig(5789, 4) == 5789

    assert round_sig_dig(24103, 1) == 20000
    assert round_sig_dig(24103, 2) == 24000
    assert round_sig_dig(24103, 3) == 24100
    assert round_sig_dig(24103, 4) == 24100
    assert round_sig_dig(24103, 5) == 24103

    assert round_sig_dig(300000, 1) == 300000

    # Floating-point values

    # Must round based on fractions, must not double-round
    assert round_sig_dig(14, 1) == 10
    assert round_sig_dig(14.0, 1) == 10
    assert round_sig_dig(14.9, 1) == 10
    assert round_sig_dig(15.0, 1) == 20
    assert round_sig_dig(15.1, 1) == 20

    assert round_sig_dig(14, 2) == 14
    assert round_sig_dig(14.0, 2) == 14
    assert round_sig_dig(14.9, 2) == 15
    assert round_sig_dig(15.0, 2) == 15
    assert round_sig_dig(15.1, 2) == 15

    # Must round to integer
    assert round_sig_dig(14, 3) == 14
    assert round_sig_dig(14.0, 3) == 14
    assert round_sig_dig(14.9, 3) == 15
    assert round_sig_dig(15.0, 3) == 15
    assert round_sig_dig(15.1, 3) == 15

    # Small integers
    assert_round_sig_dig_any_digits(0, 1)
    assert_round_sig_dig_any_digits(1, 1)
    assert_round_sig_dig_any_digits(2, 2)
    assert_round_sig_dig_any_digits(3, 3)
    assert_round_sig_dig_any_digits(4, 4)
    assert_round_sig_dig_any_digits(5, 5)
    assert_round_sig_dig_any_digits(6, 6)
    assert_round_sig_dig_any_digits(7, 7)
    assert_round_sig_dig_any_digits(8, 8)
    assert_round_sig_dig_any_digits(9, 9)
    assert_round_sig_dig_any_digits(10, 10)

    # Large values
    assert_round_sig_dig_any_digits_error(2**30)
    assert_round_sig_dig_any_digits_error(2**31)
    assert_round_sig_dig_any_digits_error(2**32)

    # the floating-point accuracy limit for this function is 2**73
    # on some machines
    assert_round_sig_dig_any_digits_error(2**62)
    assert_round_sig_dig_any_digits_error(2**63)
    assert_round_sig_dig_any_digits_error(2**64)

    # Out of range values: must round to 1
    assert_round_sig_dig_any_digits(-0.01, 1)
    assert_round_sig_dig_any_digits(-1, 1)
    assert_round_sig_dig_any_digits(-10.5, 1)
    assert_round_sig_dig_any_digits(-(2**31), 1)

    # test the transition points in the supported range
    # testing the entire range up to 1 million takes 100s
    for n in range(1, 20000):
        assert_round_sig_dig_any_digits_error(n)

    # use a step that is relatively prime, to increase the chance of
    # detecting errors
    for n in range(90000, 200000, 9):
        assert_round_sig_dig_any_digits_error(n)

    for n in range(900000, 2000000, 99):
        assert_round_sig_dig_any_digits_error(n)


def test_v3bwline_from_results_file(datadir):
    lines = datadir.readlines("results.txt")
    d = dict()
    for line in lines:
        r = Result.from_dict(json.loads(line.strip(), cls=CustomDecoder))
        fp = r.fingerprint
        if fp not in d:
            d[fp] = []
        d[fp].append(r)
    bwl, _ = V3BWLine.from_data(d, fp)
    # bw store now B, not KB
    bwl.bw = round(bwl.bw / 1000)
    assert raw_bwl_str == str(bwl)


def test_from_results_read(datadir, tmpdir, conf, args):
    results = load_result_file(str(datadir.join("results.txt")))
    expected_header = V3BWHeader(
        timestamp_l,
        earliest_bandwidth=earliest_bandwidth,
        latest_bandwidth=latest_bandwidth,
    )
    exclusion_dict = dict(
        [(k, 0) for k in HEADER_RECENT_MEASUREMENTS_EXCLUDED_KEYS]
    )
    expected_header.add_relays_excluded_counters(exclusion_dict)
    raw_bwls = [V3BWLine.from_results(results[fp])[0] for fp in results]
    # Scale BWLines using torflow method, since it's the default and BWLines
    # bandwidth is the raw bandwidth.
    expected_bwls = V3BWFile.bw_torflow_scale(raw_bwls)
    # Since the scaled lines will be less than the 60% relays in the network,
    # set under_min_report.
    expected_bwls[0].under_min_report = 1
    expected_bwls[0].vote = 0
    expected_f = V3BWFile(expected_header, expected_bwls)
    # This way is going to convert bw to KB
    v3bwfile = V3BWFile.from_results(results)
    assert str(expected_f)[1:] == str(v3bwfile)[1:]
    output = os.path.join(args.output, now_fname())
    v3bwfile.write(output)


def test_from_arg_results_write(datadir, tmpdir, conf, args):
    results = load_result_file(str(datadir.join("results.txt")))
    v3bwfile = V3BWFile.from_results(results)
    output = os.path.join(args.output, now_fname())
    v3bwfile.write(output)
    assert os.path.isfile(output)


def test_from_arg_results_write_read(datadir, tmpdir, conf, args):
    results = load_result_file(str(datadir.join("results.txt")))
    v3bwfile = V3BWFile.from_results(results)
    output = os.path.join(args.output, now_fname())
    v3bwfile.write(output)
    with open(output) as fd:
        v3bw = fd.read()
    assert v3bw == str(v3bwfile)


def test_sbws_scale(datadir):
    results = load_result_file(str(datadir.join("results.txt")))
    v3bwfile = V3BWFile.from_results(results, scaling_method=SBWS_SCALING)
    assert v3bwfile.bw_lines[0].bw == 8


def num_consensus_relays(fpath):
    return 1


# To do not have to create a consensus-cache file and set the path,
# mock the result since it only returns the number of relays.
@mock.patch.object(V3BWFile, "read_number_consensus_relays")
def test_torflow_scale(mock_consensus, datadir, tmpdir, conf):
    mock_consensus.return_value = 1
    # state_fpath = str(tmpdir.join('.sbws', 'state.dat'))
    state_fpath = conf["paths"]["state_fpath"]
    results = load_result_file(str(datadir.join("results.txt")))
    # Since v1.1.0, it'll write bw=1 if the minimum percent of measured relays
    # wasn't reached. Therefore mock the consensus number.
    # Because the consensus number is mocked, it'll try to read the state path.
    # Obtain it from conf, so that the root directory exists.
    v3bwfile = V3BWFile.from_results(
        results,
        "",
        "",
        state_fpath,
        scaling_method=TORFLOW_SCALING,
        round_digs=TORFLOW_ROUND_DIG,
    )
    assert v3bwfile.bw_lines[0].bw == 6
    v3bwfile = V3BWFile.from_results(
        results,
        "",
        "",
        state_fpath,
        scaling_method=TORFLOW_SCALING,
        torflow_cap=0.0001,
        round_digs=TORFLOW_ROUND_DIG,
    )
    # Wrong because it should be rounded after clipping
    assert v3bwfile.bw_lines[0].bw == 1
    v3bwfile = V3BWFile.from_results(
        results,
        "",
        "",
        state_fpath,
        scaling_method=TORFLOW_SCALING,
        torflow_cap=1,
        round_digs=TORFLOW_ROUND_DIG,
    )
    assert v3bwfile.bw_lines[0].bw == 123
    v3bwfile = V3BWFile.from_results(
        results,
        "",
        "",
        state_fpath,
        scaling_method=TORFLOW_SCALING,
        torflow_cap=1,
        round_digs=PROP276_ROUND_DIG,
    )
    assert v3bwfile.bw_lines[0].bw == 120


def test_torflow_scale_no_desc_bw_avg(datadir, conf, caplog):
    state_fpath = conf["paths"]["state_fpath"]
    results = load_result_file(str(datadir.join("results_no_desc_bw_avg.txt")))
    caplog.set_level(logging.DEBUG)
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath)
    assert v3bwfile.bw_lines[0].bw == 26


def test_torflow_scale_no_desc_bw_obs(datadir, conf, caplog):
    state_fpath = conf["paths"]["state_fpath"]
    results = load_result_file(str(datadir.join("results_no_desc_bw_obs.txt")))
    caplog.set_level(logging.DEBUG)
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath)
    assert v3bwfile.bw_lines[0].bw == 1


def test_torflow_scale_no_desc_bw_avg_obs(datadir, conf, caplog):
    state_fpath = conf["paths"]["state_fpath"]
    results = load_result_file(
        str(datadir.join("results_no_desc_bw_avg_obs.txt"))
    )
    caplog.set_level(logging.DEBUG)
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath)
    assert v3bwfile.bw_lines[0].bw == 1


def test_torflow_scale_no_consensus_bw(datadir, conf, caplog):
    state_fpath = conf["paths"]["state_fpath"]
    results = load_result_file(
        str(datadir.join("results_no_consensus_bw.txt"))
    )
    caplog.set_level(logging.DEBUG)
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath)
    assert v3bwfile.bw_lines[0].bw == 26


def test_torflow_scale_0_consensus_bw(datadir, conf, caplog):
    state_fpath = conf["paths"]["state_fpath"]
    results = load_result_file(str(datadir.join("results_0_consensus_bw.txt")))
    caplog.set_level(logging.DEBUG)
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath)
    assert v3bwfile.bw_lines[0].bw == 26


def test_results_away_each_other(datadir):
    min_num = 2
    secs_away = 86400  # 1d
    results = load_result_file(str(datadir.join("results_away.txt")))
    # A has 4 results, 3 are success, 2 are 1 day away, 1 is 12h away
    values = results["AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"]

    # There is one result excluded, but the relay is not excluded
    bwl, reason = V3BWLine.from_results(values, secs_away=secs_away, min_num=2)
    assert bwl.relay_recent_measurements_excluded_error_count == 1
    assert reason is None
    assert not hasattr(bwl, "vote")
    assert not hasattr(bwl, "unmeasured")

    success_results = [r for r in values if isinstance(r, ResultSuccess)]
    assert len(success_results) >= min_num
    results_away = V3BWLine.results_away_each_other(success_results, secs_away)
    assert len(results_away) == 3

    # B has 2 results, 12h away from each other
    values = results["BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB"]

    # Two measurements are excluded and there were only 2,
    # the relay is excluded
    bwl, reason = V3BWLine.from_results(values, secs_away=secs_away, min_num=2)
    assert bwl.relay_recent_measurements_excluded_near_count == 2
    assert reason == "recent_measurements_excluded_near_count"
    assert bwl.vote == 0
    assert bwl.unmeasured == 1

    success_results = [r for r in values if isinstance(r, ResultSuccess)]
    assert len(success_results) >= min_num
    results_away = V3BWLine.results_away_each_other(success_results, secs_away)
    assert not results_away

    secs_away = 43200  # 12h
    values = results["BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB"]
    success_results = [r for r in values if isinstance(r, ResultSuccess)]
    assert len(success_results) >= min_num
    results_away = V3BWLine.results_away_each_other(success_results, secs_away)
    assert len(results_away) == 2

    # C has 1 result
    values = results["CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC"]

    # There is only 1 result, the relay is excluded
    bwl, reason = V3BWLine.from_results(values, min_num=2)
    assert bwl.relay_recent_measurements_excluded_few_count == 1
    assert reason == "recent_measurements_excluded_few_count"
    assert bwl.vote == 0
    assert bwl.unmeasured == 1

    success_results = [r for r in values if isinstance(r, ResultSuccess)]
    assert len(success_results) < min_num


def test_measured_progress_stats(datadir):
    number_consensus_relays = 3
    bw_lines_raw = []
    statsd_exp = {
        "percent_eligible_relays": 100,
        "minimum_percent_eligible_relays": 60,
        "number_consensus_relays": 3,
        "minimum_number_eligible_relays": 2,
        "number_eligible_relays": 3,
    }
    min_perc_reached_before = None
    results = load_result_file(str(datadir.join("results_away.txt")))
    for fp, values in results.items():
        # log.debug("Relay fp %s", fp)
        line, _ = V3BWLine.from_results(values)
        if line is not None:
            bw_lines_raw.append(line)
    assert len(bw_lines_raw) == 3
    bw_lines = V3BWFile.bw_torflow_scale(bw_lines_raw)
    assert len(bw_lines) == 3
    statsd, success = V3BWFile.measured_progress_stats(
        len(bw_lines), number_consensus_relays, min_perc_reached_before
    )
    assert success
    assert statsd == statsd_exp
    number_consensus_relays = 6
    statsd, success = V3BWFile.measured_progress_stats(
        len(bw_lines), number_consensus_relays, min_perc_reached_before
    )
    assert not success
    statsd_exp = {
        "percent_eligible_relays": 50,
        "minimum_percent_eligible_relays": 60,
        "number_consensus_relays": 6,
        "minimum_number_eligible_relays": 4,
        "number_eligible_relays": 3,
    }
    assert statsd_exp == statsd


def test_update_progress(datadir, tmpdir):
    bw_lines_raw = []
    number_consensus_relays = 6
    state = {}
    header = V3BWHeader(str(now_unixts()))
    results = load_result_file(str(datadir.join("results_away.txt")))
    for fp, values in results.items():
        # log.debug("Relay fp %s", fp)
        line = V3BWLine.from_results(values)
        if line is not None:
            bw_lines_raw.append(line)
    bwfile = V3BWFile(header, [])
    bwfile.update_progress(
        len(bw_lines_raw), header, number_consensus_relays, state
    )
    assert header.percent_eligible_relays == "50"
    assert state.get("min_perc_reached") is None
    # Test that the headers are also included when there are enough eligible
    # relays
    number_consensus_relays = 3
    header = V3BWHeader(str(now_unixts()))
    bwfile.update_progress(
        len(bw_lines_raw), header, number_consensus_relays, state
    )
    assert state.get("min_perc_reached") == now_isodt_str()
    assert header.minimum_number_eligible_relays == "2"
    assert header.minimum_percent_eligible_relays == str(MIN_REPORT)
    assert header.number_consensus_relays == "3"
    assert header.number_eligible_relays == "3"
    assert header.percent_eligible_relays == "100"


def test_time_measure_half_network(caplog):
    header = V3BWHeader(
        timestamp_l,
        file_created=file_created,
        generator_started=generator_started,
        earliest_bandwidth=earliest_bandwidth,
    )
    header.number_consensus_relays = "6500"
    header.number_eligible_relays = "4000"
    caplog.set_level(logging.INFO)
    header.add_time_report_half_network()
    assert header.time_to_report_half_network == "70200"  # 19.5h
    expected_log = (
        "Estimated time to successfully report about all the "
        "relays in the last consensus: 39 hours."
    )  # 19.5*2
    assert caplog.records[-1].getMessage() == expected_log


@mock.patch.object(V3BWFile, "read_number_consensus_relays")
def test_set_under_min_report(mock_consensus, conf, datadir):
    # The number of relays (1) is the same as the ones in the consensus,
    # therefore there is no any relay excluded and under_min_report is not set.
    mock_consensus.return_value = 1
    state_fpath = conf["paths"]["state_fpath"]
    results = load_result_file(str(datadir.join("results.txt")))
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath)
    bwl = v3bwfile.bw_lines[0]
    assert not hasattr(bwl, "vote")
    assert not hasattr(bwl, "under_min_report")
    assert bwl.bw != 1

    # The number of relays is the same as the ones in the consensus,
    # but after filtering there's no any, under_min_report is set to 1
    # and unmeasured was also set to 1.
    # After filtering the relay is excluded because there's only 1 success
    # result and it should have at least 2 (min_num)
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath, min_num=2)
    bwl = v3bwfile.bw_lines[0]
    assert bwl.vote == 0
    assert bwl.under_min_report == 1
    assert bwl.unmeasured == 1
    assert bwl.bw == 1

    # The number of relays after scaling is than the 60% in the network,
    # therefore the relays are excluded and under_min_report is set to 1.
    mock_consensus.return_value = 3
    v3bwfile = V3BWFile.from_results(results, "", "", state_fpath)
    bwl = v3bwfile.bw_lines[0]
    assert bwl.vote == 0
    assert bwl.under_min_report == 1
    assert bwl.bw != 1


def test_generator_started(root_data_path, datadir):
    state_fpath = os.path.join(root_data_path, ".sbws/state.dat")
    # The method is correct
    assert "2020-02-29T10:00:00" == V3BWHeader.generator_started_from_file(
        state_fpath
    )
    # `results` does not matter here, using them to not have an empty list.
    results = load_result_file(str(datadir.join("results.txt")))
    header = V3BWHeader.from_results(results, "", "", state_fpath)
    # And the header is correct
    assert "2020-02-29T10:00:00" == header.generator_started


def test_recent_consensus_count(root_data_path, datadir):
    # This state has recent_consensus_count
    state_fpath = os.path.join(root_data_path, ".sbws/state.dat")
    assert "1" == V3BWHeader.consensus_count_from_file(state_fpath)
    # `results` does not matter here, using them to not have an empty list.
    results = load_result_file(str(datadir.join("results.txt")))
    header = V3BWHeader.from_results(results, "", "", state_fpath)
    assert "1" == header.recent_consensus_count


@pytest.mark.skip(reason="increment_recent_measurement_attempt() disabled")
def test_recent_measurement_attempt_count(root_data_path, datadir):
    state_fpath = os.path.join(root_data_path, ".sbws/state.dat")
    assert 15 == V3BWHeader.recent_measurement_attempt_count_from_file(
        state_fpath
    )
    # `results` does not matter here, using them to not have an empty list.
    results = load_result_file(str(datadir.join("results.txt")))
    header = V3BWHeader.from_results(results, "", "", state_fpath)
    assert "15" == header.recent_measurement_attempt_count


def test_recent_priority_list_count(root_data_path, datadir):
    # This state has recent_priority_list
    state_fpath = os.path.join(root_data_path, ".sbws/state.dat")
    assert 1 == V3BWHeader.recent_priority_list_count_from_file(state_fpath)
    # `results` does not matter here, using them to don't have an empty list.
    results = load_result_file(str(datadir.join("results.txt")))
    header = V3BWHeader.from_results(results, "", "", state_fpath)
    assert "1" == header.recent_priority_list_count


def test_recent_priority_relay_count(root_data_path, datadir):
    # This state has recent_priority_relay_count
    state_fpath = os.path.join(root_data_path, ".sbws/state.dat")
    assert 15 == V3BWHeader.recent_priority_relay_count_from_file(state_fpath)
    # `results` does not matter here, using them to don't have an empty list.
    results = load_result_file(str(datadir.join("results.txt")))
    header = V3BWHeader.from_results(results, "", "", state_fpath)
    assert "15" == header.recent_priority_relay_count


def test_relay_recent_measurement_attempt_count(root_data_path, datadir):
    results = load_result_file(str(datadir.join("results.txt")))
    for fp, values in results.items():
        line = V3BWLine.from_results(values)
    assert "2" == line[0].relay_recent_measurement_attempt_count


def test_relay_recent_priority_list_count(root_data_path, datadir):
    results = load_result_file(str(datadir.join("results.txt")))
    for fp, values in results.items():
        line = V3BWLine.from_results(values)
    assert "3" == line[0].relay_recent_priority_list_count


def test_relay_in_recent_consensus_count(root_data_path, datadir):
    results = load_result_file(str(datadir.join("results.txt")))
    for fp, values in results.items():
        line = V3BWLine.from_results(values)
    assert "3" == line[0].relay_in_recent_consensus_count



"""Util functions to manage sbws configuration files."""

import logging
import logging.config
import os
from configparser import (
    ConfigParser,
    ExtendedInterpolation,
    InterpolationMissingOptionError,
)
from string import Template
from tempfile import NamedTemporaryFile
from urllib.parse import urlparse

from sbws.globals import (
    DEFAULT_CONFIG_PATH,
    DEFAULT_LOG_CONFIG_PATH,
    SUPERVISED_RUN_DPATH,
    SUPERVISED_USER_CONFIG_PATH,
    USER_CONFIG_PATH,
)
from sbws.util.iso3166 import ISO_3166_ALPHA_2

_ALPHANUM = "abcdefghijklmnopqrstuvwxyz"
_ALPHANUM += _ALPHANUM.upper()
_ALPHANUM += "0123456789"

_SYMBOLS_NO_QUOTES = "!@#$%^&*()-_=+\\|[]{}:;/?.,<>"

_HEX = "0123456789ABCDEF"

_LOG_LEVELS = ["debug", "info", "warning", "error", "critical"]

log = logging.getLogger(__name__)


def _expand_path(path):
    """Expand path string containing shell variables and ~ constructions
    into their values. Environment variables have to have their $ escaped by
    another $. For example: $$XDG_RUNTIME_DIR/foo.bar
    """
    return os.path.expanduser(os.path.expandvars(path))


def _extend_config(conf, fname):
    """Extend ConfigParser from file configuration."""
    log.debug("Reading config file %s", fname)
    with open(fname, "rt") as fd:
        conf.read_file(fd, source=fname)
    return conf


def _get_default_config():
    """Return ConfigParser with default configuration."""
    conf = ConfigParser(
        interpolation=ExtendedInterpolation(),
        converters={"path": _expand_path},
    )
    return _extend_config(conf, DEFAULT_CONFIG_PATH)


def _obtain_user_conf_path():
    if os.environ.get("SUPERVISED") == "1":
        return SUPERVISED_USER_CONFIG_PATH
    return USER_CONFIG_PATH


def _get_user_config(args, conf=None):
    """Get user configuration.
    Search for user configuration in the default path or the path passed as
    argument and extend the configuration if they are found.
    """
    if not conf:
        conf = ConfigParser(
            interpolation=ExtendedInterpolation(),
            converters={"path": _expand_path},
        )
    else:
        assert isinstance(conf, ConfigParser)
    if args.config:
        if not os.path.isfile(args.config):
            # XXX: The logger is not configured at this stage,
            # sbws should start with a logger before reading configurations.
            # #40110: while the log is not configured, do not print, so that
            # no output is generated unless there're warnings.
            print(
                "Configuration file %s not found, using defaults."
                % args.config
            )
            return conf
        # print("Using configuration provided as argument %s" % args.config)
        return _extend_config(conf, args.config)
    user_config_path = _obtain_user_conf_path()
    if os.path.isfile(user_config_path):
        # print("Using configuration file %s" % user_config_path)
        return _extend_config(conf, user_config_path)
    log.debug("No user config found, using defaults.")
    return conf


def _get_default_logging_config(conf=None):
    """Get default logging configuration."""
    if not conf:
        conf = ConfigParser(
            interpolation=ExtendedInterpolation(),
            converters={"path": _expand_path},
        )
    else:
        assert isinstance(conf, ConfigParser)
    return _extend_config(conf, DEFAULT_LOG_CONFIG_PATH)


def get_config(args):
    """Get ConfigParser interpolating all configuration files."""
    conf = _get_default_config()
    conf = _get_default_logging_config(conf=conf)
    conf = _get_user_config(args, conf=conf)
    return conf


def _can_log_to_file(conf):
    """
    Checks all the known reasons for why we might not be able to log to a file,
    and returns whether or not we think we will be able to do so. This is
    useful because if we can't log to a file, we might want to force logging to
    stdout.

    If we can't log to file, return False and the reason. Otherwise return True
    and an empty string.
    """
    # We won't be able to get paths.log_dname from the config when we are first
    # initializing sbws because it depends on paths.sbws_home (by default).
    # If there is an issue getting this option, tell the caller that we can't
    # log to file.
    try:
        conf.getpath("paths", "log_dname")
    except InterpolationMissingOptionError as e:
        return False, e
    return True, ""


def configure_logging(args, conf):
    assert isinstance(conf, ConfigParser)
    logger = "logger_sbws"
    # Set the correct handler(s) based on [logging] options
    handlers = set()
    can_log_to_file, reason = _can_log_to_file(conf)
    if not can_log_to_file or conf.getboolean("logging", "to_stdout"):
        # always add to_stdout if we cannot log to file
        handlers.add("to_stdout")
    if can_log_to_file and conf.getboolean("logging", "to_file"):
        handlers.add("to_file")
    if conf.getboolean("logging", "to_syslog"):
        handlers.add("to_syslog")
    # Collect the handlers in the appropriate config option
    conf[logger]["handlers"] = ",".join(handlers)
    if "to_file" in handlers:
        # This is weird.
        #
        # Python's logging library expects 'args' to be a tuple ... but it has
        # to be stored as a string and it evals() the string.
        #
        # The first argument is the file name to which it should log. Set it to
        # the sbws command (like 'scanner' or 'generate') if possible, or to
        # 'sbws' failing that.
        dname = conf.getpath("paths", "log_dname")
        os.makedirs(dname, exist_ok=True)
        fname = os.path.join(dname, "{}.log".format(args.command or "sbws"))
        # The second argument is the file mode, and it should be left alone
        mode = "a"
        # The third is the maximum file size (in bytes) each log file should be
        max_bytes = conf.getint("logging", "to_file_max_bytes")
        # And the forth is the number of backups to keep
        num_backups = conf.getint("logging", "to_file_num_backups")
        # Now store those things as a string in the config. So dumb.
        conf["handler_to_file"]["args"] = str(
            (fname, mode, max_bytes, num_backups)
        )
    # Set some stuff that needs config parser's interpolation
    conf["formatter_to_file"]["format"] = conf["logging"]["to_file_format"]
    conf["formatter_to_stdout"]["format"] = conf["logging"]["to_stdout_format"]
    conf["formatter_to_syslog"]["format"] = conf["logging"]["to_syslog_format"]
    conf[logger]["level"] = conf["logging"]["level"].upper()
    conf["handler_to_file"]["level"] = conf["logging"]["to_file_level"].upper()
    conf["handler_to_stdout"]["level"] = conf["logging"][
        "to_stdout_level"
    ].upper()
    conf["handler_to_syslog"]["level"] = conf["logging"][
        "to_syslog_level"
    ].upper()
    # If there's a log_level cli argument, the user would expect that level
    # in the standard output.
    # conf['logging']['level'] sets the lower level, but it's still needed to
    # set the stdout level.
    # It also must be set up in the end, since cli arguments have higher
    # priority.
    if args.log_level:
        conf["logging"]["level"] = args.log_level.upper()
        conf["handler_to_stdout"]["level"] = conf["logging"]["level"]
    # Now we configure the standard python logging system
    with NamedTemporaryFile("w+t") as fd:
        conf.write(fd)
        fd.seek(0, 0)
        logging.config.fileConfig(fd.name)


def validate_config(conf):
    """Checks the given conf for bad values or bad combinations of values. If
    there's something wrong, returns False and a list of error messages.
    Otherwise, return True and an empty list"""
    errors = []
    errors.extend(_validate_general(conf))
    errors.extend(_validate_cleanup(conf))
    errors.extend(_validate_scanner(conf))
    errors.extend(_validate_tor(conf))
    errors.extend(_validate_paths(conf))
    errors.extend(_validate_destinations(conf))
    errors.extend(_validate_relayprioritizer(conf))
    errors.extend(_validate_logging(conf))
    return len(errors) < 1, errors


def _validate_cleanup(conf):
    errors = []
    sec = "cleanup"
    err_tmpl = Template("$sec/$key ($val): $e")
    ints = {
        "data_files_compress_after_days": {"minimum": 1, "maximum": None},
        "data_files_delete_after_days": {"minimum": 1, "maximum": None},
        "v3bw_files_compress_after_days": {"minimum": 1, "maximum": None},
        "v3bw_files_delete_after_days": {"minimum": 1, "maximum": None},
    }
    all_valid_keys = list(ints.keys())
    errors.extend(_validate_section_keys(conf, sec, all_valid_keys, err_tmpl))
    errors.extend(_validate_section_ints(conf, sec, ints, err_tmpl))
    return errors


def _validate_general(conf):
    errors = []
    sec = "general"
    err_tmpl = Template("$sec/$key ($val): $e")
    ints = {
        "data_period": {"minimum": 1, "maximum": None},
        "circuit_timeout": {"minimum": 1, "maximum": None},
    }
    floats = {
        "http_timeout": {"minimum": 0.0, "maximum": None},
    }
    bools = {
        "reset_bw_ipv4_changes": {},
        "reset_bw_ipv6_changes": {},
    }
    all_valid_keys = (
        list(ints.keys()) + list(floats.keys()) + list(bools.keys())
    )
    errors.extend(_validate_section_keys(conf, sec, all_valid_keys, err_tmpl))
    errors.extend(_validate_section_ints(conf, sec, ints, err_tmpl))
    errors.extend(_validate_section_floats(conf, sec, floats, err_tmpl))
    errors.extend(_validate_section_bools(conf, sec, bools, err_tmpl))
    return errors


def _obtain_sbws_home(conf):
    sbws_home = conf.getpath("paths", "sbws_home")
    # No need for .sbws when this is the default home
    if sbws_home == "/var/lib/sbws/.sbws":
        conf["paths"]["sbws_home"] = os.path.dirname(sbws_home)


def _obtain_run_dpath(conf):
    """Set runtime directory when sbws is run by a system service."""
    xdg = os.environ.get("XDG_RUNTIME_DIR")
    if os.environ.get("SUPERVISED") == "1":
        conf["tor"]["run_dpath"] = SUPERVISED_RUN_DPATH
    elif xdg is not None:
        conf["tor"]["run_dpath"] = os.path.join(xdg, "sbws", "tor")


def _validate_paths(conf):
    _obtain_sbws_home(conf)
    errors = []
    sec = "paths"
    err_tmpl = Template("$sec/$key ($val): $e")
    unvalidated_keys = [
        "datadir",
        "sbws_home",
        "v3bw_fname",
        "v3bw_dname",
        "state_fname",
        "log_dname",
    ]
    all_valid_keys = unvalidated_keys
    allow_missing = ["sbws_home"]
    errors.extend(
        _validate_section_keys(
            conf, sec, all_valid_keys, err_tmpl, allow_missing=allow_missing
        )
    )
    return errors


def _validate_country(conf, sec, key, err_tmpl):
    errors = []
    if conf[sec].get(key, None) is None:
        errors.append(
            err_tmpl.substitute(
                sec=sec,
                key=key,
                val=None,
                e="Missing country in configuration file.",
            )
        )
        return errors
    valid = conf[sec]["country"] in ISO_3166_ALPHA_2
    if not valid:
        errors.append(
            err_tmpl.substitute(
                sec=sec,
                key=key,
                val=conf[sec][key],
                e="Not a valid ISO 3166 alpha-2 country code.",
            )
        )
    return errors


def _validate_scanner(conf):
    errors = []
    sec = "scanner"
    err_tmpl = Template("$sec/$key ($val): $e")
    ints = {
        "num_rtts": {"minimum": 0, "maximum": 100},
        "num_downloads": {"minimum": 1, "maximum": 100},
        "initial_read_request": {"minimum": 1, "maximum": None},
        "measurement_threads": {"minimum": 1, "maximum": None},
        "min_download_size": {"minimum": 1, "maximum": None},
        "max_download_size": {"minimum": 1, "maximum": None},
    }
    floats = {
        "download_toofast": {"minimum": 0.001, "maximum": None},
        "download_min": {"minimum": 0.001, "maximum": None},
        "download_target": {"minimum": 0.001, "maximum": None},
        "download_max": {"minimum": 0.001, "maximum": None},
    }
    all_valid_keys = (
        list(ints.keys()) + list(floats.keys()) + ["nickname", "country"]
    )
    errors.extend(_validate_section_keys(conf, sec, all_valid_keys, err_tmpl))
    errors.extend(_validate_section_ints(conf, sec, ints, err_tmpl))
    errors.extend(_validate_section_floats(conf, sec, floats, err_tmpl))
    valid, error_msg = _validate_nickname(conf[sec], "nickname")
    if not valid:
        errors.append(
            err_tmpl.substitute(
                sec=sec, key="nickname", val=conf[sec]["nickname"], e=error_msg
            )
        )
    errors.extend(_validate_country(conf, sec, "country", err_tmpl))
    return errors


def _validate_tor(conf):
    _obtain_run_dpath(conf)
    errors = []
    sec = "tor"
    err_tmpl = Template("$sec/$key ($val): $e")
    unvalidated_keys = [
        "datadir",
        "run_dpath",
        "control_socket",
        "pid",
        "log",
        "external_control_port",
        "extra_lines",
    ]
    all_valid_keys = unvalidated_keys
    errors.extend(_validate_section_keys(conf, sec, all_valid_keys, err_tmpl))
    return errors


def _validate_relayprioritizer(conf):
    errors = []
    sec = "relayprioritizer"
    err_tmpl = Template("$sec/$key ($val): $e")
    ints = {
        "min_relays": {"minimum": 1, "maximum": None},
    }
    floats = {
        "fraction_relays": {"minimum": 0.0, "maximum": 1.0},
    }
    bools = {
        "measure_authorities": {},
    }
    all_valid_keys = (
        list(ints.keys()) + list(floats.keys()) + list(bools.keys())
    )
    errors.extend(_validate_section_keys(conf, sec, all_valid_keys, err_tmpl))
    errors.extend(_validate_section_ints(conf, sec, ints, err_tmpl))
    errors.extend(_validate_section_floats(conf, sec, floats, err_tmpl))
    errors.extend(_validate_section_bools(conf, sec, bools, err_tmpl))
    return errors


def _validate_logging(conf):
    errors = []
    sec = "logging"
    err_tmpl = Template("$sec/$key ($val): $e")
    enums = {
        "level": {"choices": _LOG_LEVELS},
        "to_file_level": {"choices": _LOG_LEVELS},
        "to_stdout_level": {"choices": _LOG_LEVELS},
        "to_syslog_level": {"choices": _LOG_LEVELS},
    }
    bools = {
        "to_file": {},
        "to_stdout": {},
        "to_syslog": {},
    }
    ints = {
        "to_file_max_bytes": {"minimum": 0, "maximum": None},
        "to_file_num_backups": {"minimum": 0, "maximum": None},
    }
    unvalidated = [
        "format",
        "to_file_format",
        "to_stdout_format",
        "to_syslog_format",
    ]
    all_valid_keys = (
        list(bools.keys())
        + list(enums.keys())
        + list(ints.keys())
        + unvalidated
    )
    errors.extend(_validate_section_keys(conf, sec, all_valid_keys, err_tmpl))
    errors.extend(_validate_section_bools(conf, sec, bools, err_tmpl))
    errors.extend(_validate_section_enums(conf, sec, enums, err_tmpl))
    return errors


def _validate_destinations(conf):
    errors = []
    sec = "destinations"
    section = conf[sec]
    err_tmpl = Template("$sec/$key ($val): $e")
    dest_sections = []
    for key in section.keys():
        if key == "usability_test_interval":
            value = section[key]
            valid, error_msg = _validate_int(section, key, minimum=1)
            if not valid:
                errors.append(
                    err_tmpl.substitute(
                        sec=sec, key=key, val=value, e=error_msg
                    )
                )
            continue
        value = section[key]
        valid, error_msg = _validate_boolean(section, key)
        if not valid:
            errors.append(
                err_tmpl.substitute(sec=sec, key=key, val=value, e=error_msg)
            )
            continue
        assert valid
        if section.getboolean(key):
            dest_sections.append("{}.{}".format(sec, key))
    urls = {
        "url": {},
    }
    all_valid_keys = list(urls.keys()) + [
        "verify",
        "country",
        "max_num_failures",
    ]
    for sec in dest_sections:
        if sec not in conf:
            errors.append(
                "{} is an enabled destination but is not a "
                "section in the config".format(sec)
            )
            continue
        errors.extend(
            _validate_section_keys(
                conf,
                sec,
                all_valid_keys,
                err_tmpl,
                allow_missing=["verify", "max_num_failures"],
            )
        )
        errors.extend(_validate_section_urls(conf, sec, urls, err_tmpl))
        errors.extend(_validate_country(conf, sec, "country", err_tmpl))
    return errors


def _validate_section_keys(conf, sec, keys, tmpl, allow_missing=None):
    if allow_missing is None:
        allow_missing = []
    errors = []
    section = conf[sec]
    # Find keys that exist in the user's config that are not known
    for key in section:
        if key not in keys:
            errors.append(
                tmpl.substitute(
                    sec=sec, key=key, val=section[key], e="Unknown key"
                )
            )
    # Find keys that don't exist in the user's config that should
    for key in keys:
        if key not in section and key not in allow_missing:
            errors.append(
                tmpl.substitute(
                    sec=sec, key=key, val="[NOT SET]", e="Missing key"
                )
            )
    return errors


def _validate_section_ints(conf, sec, ints, tmpl):
    errors = []
    section = conf[sec]
    for key in ints:
        valid, error = _validate_int(
            section,
            key,
            minimum=ints[key]["minimum"],
            maximum=ints[key]["maximum"],
        )
        if not valid:
            errors.append(
                tmpl.substitute(sec=sec, key=key, val=section[key], e=error)
            )
    return errors


def _validate_section_floats(conf, sec, floats, tmpl):
    errors = []
    section = conf[sec]
    for key in floats:
        valid, error = _validate_float(
            section,
            key,
            minimum=floats[key]["minimum"],
            maximum=floats[key]["maximum"],
        )
        if not valid:
            errors.append(
                tmpl.substitute(sec=sec, key=key, val=section[key], e=error)
            )
    return errors


def _validate_section_hosts(conf, sec, hosts, tmpl):
    errors = []
    section = conf[sec]
    for key in hosts:
        valid, error = _validate_host(section, key)
        if not valid:
            errors.append(
                tmpl.substitute(sec=sec, key=key, val=section[key], e=error)
            )
    return errors


def _validate_section_ports(conf, sec, ports, tmpl):
    errors = []
    section = conf[sec]
    for key in ports:
        valid, error = _validate_int(section, key, minimum=1, maximum=2**16)
        if not valid:
            errors.append(
                tmpl.substitute(
                    sec=sec,
                    key=key,
                    val=section[key],
                    e="Not a valid port ({})".format(error),
                )
            )
    return errors


def _validate_section_bools(conf, sec, bools, tmpl):
    errors = []
    section = conf[sec]
    for key in bools:
        valid, error = _validate_boolean(section, key)
        if not valid:
            errors.append(
                tmpl.substitute(
                    sec=sec,
                    key=key,
                    val=section[key],
                    e="Not a valid boolean string ({})".format(error),
                )
            )
    return errors


def _validate_section_fingerprints(conf, sec, fps, tmpl):
    errors = []
    section = conf[sec]
    for key in fps:
        valid, error = _validate_fingerprint(section, key)
        if not valid:
            errors.append(
                tmpl.substitute(
                    sec=sec,
                    key=key,
                    val=section[key],
                    e="Not a valid fingerprint ({})".format(error),
                )
            )
    return errors


def _validate_section_urls(conf, sec, urls, tmpl):
    errors = []
    section = conf[sec]
    for key in urls:
        valid, error = _validate_url(section, key)
        if not valid:
            errors.append(
                tmpl.substitute(
                    sec=sec,
                    key=key,
                    val=section[key],
                    e="Not a valid url ({})".format(error),
                )
            )
    return errors


def _validate_section_enums(conf, sec, enums, tmpl):
    errors = []
    section = conf[sec]
    for key in enums:
        choices = enums[key]["choices"]
        valid, error = _validate_enum(section, key, choices)
        if not valid:
            errors.append(
                tmpl.substitute(
                    sec=sec,
                    key=key,
                    val=section[key],
                    e="Not a valid enum choice ({})".format(
                        ", ".join(choices)
                    ),
                )
            )
    return errors


def _validate_enum(section, key, choices):
    value = section[key]
    if value not in choices:
        return False, "{} not in allowed choices: {}".format(
            value, ", ".join(choices)
        )
    return True, ""


def _validate_url(section, key):
    value = section[key]
    url = urlparse(value)
    if not url.netloc:
        return False, "Does not appear to contain a hostname"
    # It should be possible to have an URL that starts by http:// that uses
    # TLS,but python requests is just checking the scheme starts by https
    # when verifying certificate:
    # https://github.com/requests/requests/blob/master/requests/adapters.py#L215  # noqa
    # When the scheme is https but the protocol is not TLS, requests will hang.
    if url.scheme != "https" and not url.netloc.startswith("127.0.0.1"):
        return False, "URL scheme must be HTTPS (except for the test server)"
    return True, ""


def _validate_int(section, key, minimum=None, maximum=None):
    try:
        value = section.getint(key)
    except ValueError as e:
        return False, e
    if minimum is not None:
        assert isinstance(minimum, int)
        if value < minimum:
            return False, "Cannot be less than {}".format(minimum)
    if maximum is not None:
        assert isinstance(maximum, int)
        if value > maximum:
            return False, "Cannot be greater than {}".format(maximum)
    return True, ""


def _validate_boolean(section, key):
    try:
        section.getboolean(key)
    except ValueError as e:
        return False, e
    return True, ""


def _validate_float(section, key, minimum=None, maximum=None):
    try:
        value = section.getfloat(key)
    except ValueError as e:
        return False, e
    if minimum is not None:
        assert isinstance(minimum, float)
        if value < minimum:
            return False, "Cannot be less than {}".format(minimum)
    if maximum is not None:
        assert isinstance(maximum, float)
        if value > maximum:
            return False, "Cannot be greater than {}".format(maximum)
    return True, ""


def _validate_host(section, key):
    # XXX: Implement this
    return True, ""


def _validate_fingerprint(section, key):
    alphabet = _HEX
    length = 40
    return _validate_string(
        section, key, min_len=length, max_len=length, alphabet=alphabet
    )


def _validate_nickname(section, key):
    alphabet = _ALPHANUM + _SYMBOLS_NO_QUOTES
    min_len = 1
    max_len = 32
    return _validate_string(
        section, key, min_len=min_len, max_len=max_len, alphabet=alphabet
    )


def _validate_string(
    section, key, min_len=None, max_len=None, alphabet=None, starts_with=None
):
    s = section[key]
    if min_len is not None and len(s) < min_len:
        return False, "{} is below minimum allowed length {}".format(
            len(s), min_len
        )
    if max_len is not None and len(s) > max_len:
        return False, "{} is above maximum allowed length {}".format(
            len(s), max_len
        )
    if alphabet is not None:
        for i, c in enumerate(s):
            if c not in alphabet:
                return (
                    False,
                    "Letter {} at position {} is not in allowed "
                    'characters "{}"'.format(c, i, alphabet),
                )
    if starts_with is not None:
        if not s.startswith(starts_with):
            return False, "{} does not start with {}".format(s, starts_with)
    return True, ""



import sbws.util.stem as stem_utils


def test_launch_and_okay(persistent_launch_tor):
    cont = persistent_launch_tor
    assert stem_utils.is_bootstrapped(cont)


def test_set_torrc_runtime_option_succesful(persistent_launch_tor):
    controller = persistent_launch_tor
    runtime_options = controller.get_conf_map(["__LeaveStreamsUnattached"])
    assert runtime_options == {"__LeaveStreamsUnattached": ["1"]}


def test_set_torrc_runtime_invalidrequest_option_fail(persistent_launch_tor):
    controller = persistent_launch_tor
    try:
        controller.set_conf("ControlSocket", "/tmp/dummy")
    except stem_utils.InvalidRequest as e:
        assert "Unable to set option" in e.message


def test_set_torrc_options_can_fail_option_fail(persistent_launch_tor):
    controller = persistent_launch_tor
    try:
        controller.set_conf("BadOption", "0")
    except stem_utils.InvalidArguments as e:
        assert "Unknown option" in e.message



"""pytest configuration for unit tests."""
import argparse
import os
from datetime import datetime

import pytest

from sbws.globals import RESULT_VERSION
from sbws.lib.resultdump import (
    Result,
    ResultErrorStream,
    ResultSuccess,
    write_result_to_datadir,
)
from sbws.util.config import _get_default_config

TIME1 = 1529232277.9028733
TIME2 = datetime.utcnow().timestamp()
FP1 = "A" * 40
FP2 = "B" * 40
ED25519 = "g+Shk00y9Md0hg1S6ptnuc/wWKbADBgdjT0Kg+TSF3s"
CIRC12 = [FP1, FP2]
CIRC21 = [FP2, FP1]
DEST_URL = "http://example.com/sbws.bin"
NICK1 = "A"
NICK2 = "B"
IP1 = "169.254.100.1"
IP2 = "169.254.100.2"
RTTS1 = [
    0.2943556308746338,
    0.2885427474975586,
    0.2802879810333252,
    0.28124427795410156,
    0.2827129364013672,
    0.2901294231414795,
    0.2784590721130371,
    0.2838578224182129,
    0.2842121124267578,
    0.28656768798828125,
]
RTTS2 = [
    0.9097037315368652,
    0.9293286800384521,
    2.3764255046844482,
    0.869133710861206,
    0.8188621997833252,
    0.9046516418457031,
    1.3477752208709717,
    0.8118226528167725,
    0.8821918964385986,
    0.8746812343597412,
]
RTTS3 = [
    0.510988712310791,
    0.4889242649078369,
    0.5003941059112549,
    0.49333715438842773,
    0.5000274181365967,
    0.5426476001739502,
    0.5190870761871338,
    0.4908745288848877,
    0.5516133308410645,
    0.4995298385620117,
]
DOWNLOADS1 = [
    {"duration": 28.01000952720642, "amount": 25916542},
    {"duration": 28.203476428985596, "amount": 25916542},
    {"duration": 27.897520780563354, "amount": 25916542},
    {"duration": 29.330559492111206, "amount": 25916542},
    {"duration": 27.93175745010376, "amount": 25916542},
]
DOWNLOADS2 = [
    {"duration": 23.68175435066223, "amount": 81920},
    {"duration": 27.667736768722534, "amount": 81920},
    {"duration": 31.022956371307373, "amount": 81920},
    {"duration": 33.020694971084595, "amount": 81920},
    {"duration": 33.59471535682678, "amount": 81920},
]
DOWNLOADS3 = [
    {"duration": 30.008347988128662, "amount": 644411},
    {"duration": 30.73241639137268, "amount": 644411},
    {"duration": 31.845987796783447, "amount": 644411},
    {"duration": 29.703084230422974, "amount": 644411},
    {"duration": 30.438726663589478, "amount": 644411},
]
SCANNER = "test"
AVG_BW = 966080
BUR_BW = 1048576
OBS_BW = 524288
BW = 600000
UNMEASURED = False

now = datetime.utcnow()

RELAY1 = Result.Relay(
    FP1,
    NICK1,
    IP1,
    ED25519,
    average_bandwidth=AVG_BW,
    burst_bandwidth=BUR_BW,
    observed_bandwidth=OBS_BW,
    consensus_bandwidth=BW,
    consensus_bandwidth_is_unmeasured=UNMEASURED,
    relay_in_recent_consensus=[now],
    relay_recent_measurement_attempt=[now],
    relay_recent_priority_list=[now],
)
RELAY2 = Result.Relay(FP2, NICK2, IP2, ED25519)

RESULT = Result(RELAY1, CIRC12, DEST_URL, SCANNER, t=TIME1)
RESULT_SUCCESS1 = ResultSuccess(
    RTTS1, DOWNLOADS1, RELAY1, CIRC12, DEST_URL, SCANNER, t=TIME1
)
RESULT_SUCCESS2 = ResultSuccess(
    RTTS2, DOWNLOADS2, RELAY2, CIRC21, DEST_URL, SCANNER, t=TIME2
)
RESULT_ERROR_STREAM = ResultErrorStream(
    RELAY1, CIRC12, DEST_URL, SCANNER, t=TIME1, msg="Something bad"
)

RESULTDICT_IP_CHANGED = {FP1: [RESULT_SUCCESS1, RESULT_SUCCESS2]}
RESULTDICT_IP_NOT_CHANGED = {FP1: [RESULT_SUCCESS1, RESULT_SUCCESS1]}

RELAY_DICT = {
    "fingerprint": FP1,
    "address": IP1,
    "nickname": NICK1,
    "master_key_ed25519": ED25519,
    "relay_average_bandwidth": AVG_BW,
    "relay_burst_bandwidth": BUR_BW,
    "relay_observed_bandwidth": OBS_BW,
    "consensus_bandwidth": BW,
    "consensus_bandwidth_is_unmeasured": UNMEASURED,
    "relay_in_recent_consensus": [now],
    "relay_recent_measurement_attempt": [now],
    "relay_recent_priority_list": [now],
}

BASE_RESULT_NO_RELAY_DICT = {
    "dest_url": DEST_URL,
    "time": TIME1,
    "circ": CIRC12,
    "version": RESULT_VERSION,
    "scanner": SCANNER,
}

BASE_RESULT_DICT = RELAY_DICT.copy()
BASE_RESULT_DICT.update(BASE_RESULT_NO_RELAY_DICT)

RESULT_ERROR_STREAM_DICT = BASE_RESULT_DICT.copy()
RESULT_ERROR_STREAM_DICT.update(
    {
        "type": "error-stream",
        "msg": "Something bad",
    }
)

RESULT_SUCCESS_DICT = BASE_RESULT_DICT.copy()
RESULT_SUCCESS_DICT.update(
    {
        "rtts": RTTS1,
        "type": "success",
        "downloads": DOWNLOADS1,
    }
)


class _PseudoArguments(argparse.Namespace):

    """Just enough of the argparse.Namespace (what you get when you do
    args = parser.parse_args()) to make get_config() happy

    >>> args = _PseudoArguments(directory='/home/matt/.sbws')
    >>> args.directory
    '/home/matt/.sbws'

    """

    def __init__(self, **kw):
        for key in kw:
            setattr(self, key, kw[key])


@pytest.fixture(scope="function")
def tmpdir(tmpdir_factory, request):
    """Create a tmp dir for the tests"""
    base = str(hash(request.node.nodeid))[:3]
    bn = tmpdir_factory.mktemp(base)
    return bn


@pytest.fixture(scope="function")
def sbwshome_empty(tmpdir):
    """Create sbws home inside of the tests tmp dir without initializing."""
    home = tmpdir.join(".sbws")
    os.makedirs(home.strpath, exist_ok=True)
    return home.strpath


@pytest.fixture(scope="function")
def sbwshome_only_datadir(sbwshome_empty):
    """Create sbws home inside of the tests tmp dir with only datadir."""
    os.makedirs(os.path.join(sbwshome_empty, "datadir"), exist_ok=True)
    return sbwshome_empty


@pytest.fixture(scope="function")
def test_config_path(tmpdir):
    config = tmpdir.join(".sbws.ini")
    return config


@pytest.fixture(scope="function")
def args(sbwshome_empty, parser, test_config_path):
    """Args with sbws home in the tests tmp dir."""
    args = _PseudoArguments(
        config=test_config_path,
        output=sbwshome_empty,
        scale=False,
        log_level="debug",
    )
    return args


@pytest.fixture(scope="function")
def conf(sbwshome_empty, tmpdir):
    """Default configuration with sbws home in the tmp test dir."""
    conf = _get_default_config()
    conf["paths"]["sbws_home"] = sbwshome_empty
    conf["paths"]["state_fpath"] = str(tmpdir.join(".sbws", "state.dat"))
    conf["destinations"]["local"] = "on"
    conf["destinations.local"] = {
        "url": "http://127.0.0.1:28888/sbws.bin",
        "verify": False,
        "country": "ZZ",
    }

    return conf


@pytest.fixture(scope="function")
def conf_results(sbwshome_success_result_two_relays, conf):
    """Minimal configuration having a datadir

    So that `ResultDump` does not raise AssertionError.

    """
    conf["paths"]["sbws_home"] = sbwshome_success_result_two_relays
    return conf


@pytest.fixture()
def result():
    return RESULT


@pytest.fixture()
def result_success():
    return RESULT_SUCCESS1


@pytest.fixture()
def result_success_dict():
    return RESULT_SUCCESS_DICT


@pytest.fixture()
def result_error_stream_dict():
    return RESULT_ERROR_STREAM_DICT


@pytest.fixture()
def result_error_stream():
    return RESULT_ERROR_STREAM


@pytest.fixture()
def resultdict_ip_changed():
    return RESULTDICT_IP_CHANGED


@pytest.fixture()
def resultdict_ip_not_changed():
    return RESULTDICT_IP_NOT_CHANGED


@pytest.fixture()
def resultdict_ip_changed_trimmed():
    return {FP1: [RESULT_SUCCESS2]}


@pytest.fixture(scope="function")
def sbwshome_error_result(sbwshome_only_datadir, conf):
    """
    Creates an ~/.sbws with a single fresh ResultError in it
    """
    dd = conf.getpath("paths", "datadir")
    write_result_to_datadir(RESULT_ERROR_STREAM, dd)
    return sbwshome_only_datadir


@pytest.fixture(scope="function")
def sbwshome_success_result(sbwshome_only_datadir, conf):
    """
    Creates an ~/.sbws with a single fresh ResultSuccess in it
    """
    dd = conf.getpath("paths", "datadir")
    write_result_to_datadir(RESULT_SUCCESS1, dd)
    return sbwshome_only_datadir


@pytest.fixture(scope="function")
def sbwshome_success_result_one_relay(sbwshome_only_datadir, conf):
    """
    Creates an ~/.sbws with a a couple of fresh ResultSuccess for one relay
    """
    dd = conf.getpath("paths", "datadir")
    write_result_to_datadir(RESULT_SUCCESS1, dd)
    write_result_to_datadir(RESULT_SUCCESS1, dd)
    return sbwshome_only_datadir


@pytest.fixture(scope="function")
def sbwshome_success_result_two_relays(sbwshome_only_datadir, conf):
    """
    Creates an ~/.sbws with a a couple of fresh ResultSuccess for a couple or
    relays
    """
    dd = conf.getpath("paths", "datadir")
    write_result_to_datadir(RESULT_SUCCESS1, dd)
    write_result_to_datadir(RESULT_SUCCESS1, dd)
    write_result_to_datadir(RESULT_SUCCESS2, dd)
    write_result_to_datadir(RESULT_SUCCESS2, dd)
    return sbwshome_only_datadir


@pytest.fixture(scope="function")
def end_event():
    import threading

    return threading.Event()


@pytest.fixture(scope="function")
def rd(args, conf_results):
    from sbws.lib.resultdump import ResultDump

    return ResultDump(args, conf_results)



"""Integration tests for circutibuilder.py"""
import random


def test_build_circuit(cb, rl):
    # Path is empty
    path = []
    circuit_id, _ = cb.build_circuit(path)
    assert not circuit_id
    # Valid path, not valid exit
    exits = rl.exits_not_bad_allowing_port(port=443)
    # See https://gitlab.torproject.org/tpo/core/chutney/-/issues/40013:
    # Work around to get supposed non-exits because chutney is putting Exit
    # flag to all relays
    non_exits = list(set(rl.exits).difference(set(exits)))
    entry = random.choice(non_exits)
    # Because in chutney all relays are exits, we can't test using a non-exit
    # as 2nd hop.
    # Valid path and relays
    exit_relay = random.choice(exits)
    path = [entry.fingerprint, exit_relay.fingerprint]
    circuit_id, _ = cb.build_circuit(path)
    assert circuit_id


def test_build_circuit_with_exit_flowctrl(is_cc_tor_version, cb, rl):
    if not is_cc_tor_version:
        import pytest

        pytest.skip("This test can't be run with this tor version")
        return
    rl.consensus_params_dict = {"cc_alg": 2, "bwscanner_cc": 1}
    # Path is empty
    path = []
    circuit_id, _ = cb.build_circuit(path)
    assert not circuit_id
    # Valid path, not valid exit
    exits = rl.exits_with_2_in_flowctrl(port=443)
    # Valid path and relays
    exit_relay = random.choice(exits)
    # See https://gitlab.torproject.org/tpo/core/chutney/-/issues/40013:
    # Work around to get supposed non-exits because chutney is putting Exit
    # flag to all relays
    entry = random.choice([e for e in exits if e != exit_relay])
    path = [entry.fingerprint, exit_relay.fingerprint]
    circuit_id, _ = cb.build_circuit(path)
    assert circuit_id



import json
import logging
import os
import time
from datetime import datetime, timedelta
from enum import Enum
from glob import glob
from queue import Empty, Queue
from threading import RLock, Thread

from sbws.globals import RESULT_VERSION, fail_hard
from sbws.lib.relaylist import Relay
from sbws.util.filelock import DirectoryLock
from sbws.util.json import CustomDecoder, CustomEncoder

from .. import settings

log = logging.getLogger(__name__)


def merge_result_dicts(d1, d2):
    """
    Given two dictionaries that contain Result data, merge them.  Result
    dictionaries have keys of relay fingerprints and values of lists of results
    for those relays.
    """
    for key in d2:
        if key not in d1:
            d1[key] = []
        d1[key].extend(d2[key])
    return d1


def load_result_file(fname, success_only=False):
    """Reads in all lines from the given file, and parses them into Result
    structures (or subclasses of Result). Optionally only keeps ResultSuccess.
    Returns all kept Results as a result dictionary. This function does not
    care about the age of the results"""
    assert os.path.isfile(fname)
    d = {}
    num_total = 0
    num_ignored = 0
    with DirectoryLock(os.path.dirname(fname)):
        with open(fname, "rt") as fd:
            for line in fd:
                num_total += 1
                try:
                    r = Result.from_dict(
                        json.loads(line.strip(), cls=CustomDecoder)
                    )
                except json.decoder.JSONDecodeError:
                    log.warning("Could not decode result %s", line.strip())
                    r = None
                if r is None:
                    num_ignored += 1
                    continue
                if success_only and isinstance(r, ResultError):
                    continue
                fp = r.fingerprint
                if fp not in d:
                    d[fp] = []
                d[fp].append(r)
    num_kept = sum([len(d[fp]) for fp in d])
    log.debug("Keeping %d/%d read lines from %s", num_kept, num_total, fname)
    if num_ignored > 0:
        log.warning(
            "Had to ignore %d results due to not knowing how to "
            "parse them.",
            num_ignored,
        )
    return d


def trim_results(fresh_days, result_dict):
    """Given a result dictionary, remove all Results that are no longer valid
    and return the new dictionary"""
    assert isinstance(fresh_days, int)
    assert isinstance(result_dict, dict)
    data_period = fresh_days * 24 * 60 * 60
    oldest_allowed = time.time() - data_period
    out_results = {}
    for fp in result_dict:
        for result in result_dict[fp]:
            if result.time >= oldest_allowed:
                if fp not in out_results:
                    out_results[fp] = []
                out_results[fp].append(result)
    num_in = sum([len(result_dict[fp]) for fp in result_dict])
    num_out = sum([len(out_results[fp]) for fp in out_results])
    log.debug("Keeping %d/%d results after removing old ones", num_out, num_in)
    return out_results


def trim_results_ip_changed(
    result_dict, on_changed_ipv4=False, on_changed_ipv6=False
):
    """When there are results for the same relay with different IPs,
    create a new results' dictionary without that relay's results using an
    older IP.

    :param dict result_dict: a dictionary of results
    :param bool on_changed_ipv4: whether to trim the results when a relay's
        IPv4 changes
    :param bool on_changed_ipv6: whether to trim the results when a relay's
        IPv6 changes
    :returns: a new results dictionary
    """
    assert isinstance(result_dict, dict)
    new_results_dict = {}
    if on_changed_ipv4 is True:
        for fp in result_dict.keys():
            results = result_dict[fp]
            # find if the results for a relay have more than one ipv4
            # address
            ipv4s = set([result.address for result in results])
            if len(ipv4s) > 1:
                # keep only the results for the last ip used
                # probably we should not just discard all the results for
                # a relay that change address
                ordered_results = sorted(results, key=lambda r: r.time)
                latest_address = ordered_results[-1].address
                last_ip_results = [
                    result
                    for result in results
                    if result.address == latest_address
                ]
                new_results_dict[fp] = last_ip_results
            else:
                new_results_dict[fp] = results
        return new_results_dict
    if on_changed_ipv6 is True:
        log.warning(
            "Resetting bandwidth results when IPv6 changes,"
            " is not yet implemented."
        )
    return result_dict


def load_recent_results_in_datadir(
    fresh_days,
    datadir,
    success_only=False,
    on_changed_ipv4=False,
    on_changed_ipv6=False,
):
    """Given a data directory, read all results files in it that could have
    results in them that are still valid. Trim them, and return the valid
    Results as a list"""
    assert isinstance(fresh_days, int)
    assert os.path.isdir(datadir)
    # Inform the results are being loaded, since it takes some seconds.
    log.info("Reading and processing previous measurements.")
    results = {}
    today = datetime.utcfromtimestamp(time.time())
    data_period = fresh_days + 2
    oldest_day = today - timedelta(days=data_period)
    working_day = oldest_day
    while working_day <= today:
        # Cannot use ** and recursive=True in glob() because we support 3.4
        # So instead settle on finding files in the datadir and one
        # subdirectory below the datadir that fit the form of YYYY-MM-DD*.txt
        d = working_day.date()
        patterns = [
            os.path.join(datadir, "{}*.txt".format(d)),
            os.path.join(datadir, "*", "{}*.txt".format(d)),
        ]
        for pattern in patterns:
            for fname in glob(pattern):
                new_results = load_result_file(
                    fname, success_only=success_only
                )
                results = merge_result_dicts(results, new_results)
        working_day += timedelta(days=1)
    results = trim_results(fresh_days, results)
    # in time fresh days is possible that a relay changed ip,
    # if that's the case, keep only the results for the last ip
    results = trim_results_ip_changed(
        results, on_changed_ipv4, on_changed_ipv6
    )
    num_res = sum([len(results[fp]) for fp in results])
    if num_res == 0:
        log.warning(
            "Results files that are valid not found. "
            "Probably sbws scanner was not run first or "
            "it ran more than %d days ago or "
            "it was using a different datadir than %s.",
            data_period,
            datadir,
        )
    return results


def write_result_to_datadir(result, datadir):
    """Can be called from any thread"""
    assert isinstance(result, Result)
    assert os.path.isdir(datadir)
    dt = datetime.utcfromtimestamp(result.time)
    ext = ".txt"
    result_fname = os.path.join(datadir, "{}{}".format(dt.date(), ext))
    with DirectoryLock(datadir):
        log.debug("Writing a result to %s", result_fname)
        with open(result_fname, "at") as fd:
            fd.write("{}\n".format(str(result)))


class _StrEnum(str, Enum):
    pass


class _ResultType(_StrEnum):
    Success = "success"
    Error = "error-misc"
    ErrorCircuit = "error-circ"
    ErrorStream = "error-stream"
    ErrorAuth = "error-auth"
    # When it can not be found a second relay suitable to measure a relay.
    # It is used in ``ResultErrorSecondRelay``.
    ErrorSecondRelay = "error-second-relay"
    # When there is not a working destination Web Server.
    # It is used in ``ResultErrorDestionation``.
    ErrorDestination = "error-destination"


class Result:
    """A bandwidth measurement for a relay.

    It re-implements :class:`~sbws.lib.relaylist.Relay` as a inner class.
    """

    class Relay:
        """A Tor relay.

        It re-implements :class:`~sbws.lib.relaylist.Relay`
        with the attributes needed.

        .. note:: in a future refactor it would be simpler if a ``Relay`` has
           measurements and a measurement has a relay,
           instead of every measurement re-implementing ``Relay``.
        """

        def __init__(
            self,
            fingerprint,
            nickname,
            address,
            master_key_ed25519,
            average_bandwidth=None,
            burst_bandwidth=None,
            observed_bandwidth=None,
            consensus_bandwidth=None,
            consensus_bandwidth_is_unmeasured=None,
            # Counters to be stored by relay and not per measurement,
            # since the measurements might fail.
            relay_in_recent_consensus=None,
            relay_recent_measurement_attempt=None,
            relay_recent_priority_list=None,
        ):
            """
            Initializes a ``Result.Relay``.

            .. note:: in a future refactor the attributes should be dynamic
               to easy adding/removing them.
               They are shared by  :class:`~sbws.lib.relaylist.Relay` and
               :class:`~sbws.lib.v3bwfile.V3BWLine` and there should not be
               repeated in every class.
            """
            self.fingerprint = fingerprint
            self.nickname = nickname
            self.address = address
            self.master_key_ed25519 = master_key_ed25519
            self.average_bandwidth = average_bandwidth
            self.burst_bandwidth = burst_bandwidth
            self.observed_bandwidth = observed_bandwidth
            self.consensus_bandwidth = consensus_bandwidth
            self.consensus_bandwidth_is_unmeasured = (
                consensus_bandwidth_is_unmeasured
            )
            self.relay_in_recent_consensus = relay_in_recent_consensus
            self.relay_recent_measurement_attempt = (
                relay_recent_measurement_attempt
            )
            self.relay_recent_priority_list = relay_recent_priority_list

    def __init__(self, relay, circ, dest_url, scanner_nick, t=None):
        """
        Initializes the measurement and the relay with all the relay
        attributes.
        """
        self._relay = Result.Relay(
            relay.fingerprint,
            relay.nickname,
            relay.address,
            relay.master_key_ed25519,
            relay.average_bandwidth,
            relay.burst_bandwidth,
            relay.observed_bandwidth,
            relay.consensus_bandwidth,
            relay.consensus_bandwidth_is_unmeasured,
            relay.relay_in_recent_consensus,
            relay.relay_recent_measurement_attempt,
            relay.relay_recent_priority_list,
        )
        self._circ = circ
        self._dest_url = dest_url
        self._scanner = scanner_nick
        self._time = time.time() if t is None else t

    @property
    def type(self):
        raise NotImplementedError()

    @property
    def relay_average_bandwidth(self):
        return self._relay.average_bandwidth

    @property
    def relay_burst_bandwidth(self):
        return self._relay.burst_bandwidth

    @property
    def relay_observed_bandwidth(self):
        return self._relay.observed_bandwidth

    @property
    def consensus_bandwidth(self):
        return self._relay.consensus_bandwidth

    @property
    def consensus_bandwidth_is_unmeasured(self):
        return self._relay.consensus_bandwidth_is_unmeasured

    @property
    def fingerprint(self):
        return self._relay.fingerprint

    @property
    def nickname(self):
        return self._relay.nickname

    @property
    def address(self):
        return self._relay.address

    @property
    def master_key_ed25519(self):
        return self._relay.master_key_ed25519

    @property
    def relay_in_recent_consensus(self):
        """Number of times the relay was in a consensus."""
        return self._relay.relay_in_recent_consensus

    @property
    def relay_recent_measurement_attempt(self):
        """Returns the relay recent measurements attempts.

        It is initialized in :class:`~sbws.lib.relaylist.Relay` and
        incremented in :func:`~sbws.core.scanner.main_loop`.
        """
        return self._relay.relay_recent_measurement_attempt

    @property
    def relay_recent_priority_list(self):
        """Returns the relay recent "prioritization"s to be measured.

        It is initialized in :class:`~sbws.lib.relaylist.Relay` and
        incremented in :func:`~sbws.core.scanner.main_loop`.
        """
        return self._relay.relay_recent_priority_list

    @property
    def circ(self):
        return self._circ

    @property
    def dest_url(self):
        return self._dest_url

    @property
    def scanner(self):
        return self._scanner

    @property
    def time(self):
        return self._time

    @property
    def version(self):
        return RESULT_VERSION

    def to_dict(self):
        return {
            "fingerprint": self.fingerprint,
            "nickname": self.nickname,
            "address": self.address,
            "master_key_ed25519": self.master_key_ed25519,
            "circ": self.circ,
            "dest_url": self.dest_url,
            "time": self.time,
            "type": self.type,
            "scanner": self.scanner,
            "version": self.version,
            "relay_in_recent_consensus": self.relay_in_recent_consensus,
            "relay_recent_measurement_attempt": self.relay_recent_measurement_attempt,  # noqa
            "relay_recent_priority_list": self.relay_recent_priority_list,
        }

    @staticmethod
    def from_dict(d):
        """
        Returns a :class:`~sbws.lib.resultdump.Result` subclass from a
        dictionary.

        Returns None if the ``version`` attribute is not
        :const:`~sbws.globals.RESULT_VERSION`

        It raises ``NotImplementedError`` when the dictionary ``type`` can not
        be parsed.

        .. note:: in a future refactor, the conversions to/from
           object-dictionary will be simpler using ``setattr`` and ``__dict__``

           ``version`` is not being used and should be removed.
        """
        assert "version" in d
        if d["version"] != RESULT_VERSION:
            return None
        assert "type" in d
        if d["type"] == _ResultType.Success.value:
            return ResultSuccess.from_dict(d)
        elif d["type"] == _ResultType.Error.value:
            return ResultError.from_dict(d)
        elif d["type"] == _ResultType.ErrorCircuit.value:
            return ResultErrorCircuit.from_dict(d)
        elif d["type"] == _ResultType.ErrorStream.value:
            return ResultErrorStream.from_dict(d)
        elif d["type"] == _ResultType.ErrorAuth.value:
            return ResultErrorAuth.from_dict(d)
        elif d["type"] == _ResultType.ErrorSecondRelay.value:
            return ResultErrorSecondRelay.from_dict(d)
        elif d["type"] == _ResultType.ErrorDestination.value:
            return ResultErrorDestination.from_dict(d)
        else:
            raise NotImplementedError(
                "Unknown result type {}".format(d["type"])
            )

    def __str__(self):
        return json.dumps(self.to_dict(), cls=CustomEncoder)


class ResultError(Result):
    def __init__(self, *a, msg=None, **kw):
        super().__init__(*a, **kw)
        self._msg = msg

    @property
    def type(self):
        return _ResultType.Error

    @property
    def freshness_reduction_factor(self):
        """
        When the RelayPrioritizer encounters this Result, how much should it
        adjust its freshness? (See RelayPrioritizer.best_priority() for more
        information about "freshness")

        A higher factor makes the freshness lower (making the Result seem
        older). A lower freshness leads to the relay having better priority,
        and better priority means it will be measured again sooner.

        The value 0.5 was chosen somewhat arbitrarily, but a few weeks of live
        network testing verifies that sbws is still able to perform useful
        measurements in a reasonable amount of time.
        """
        return 0.5

    @property
    def msg(self):
        return self._msg

    @staticmethod
    def from_dict(d):
        assert isinstance(d, dict)
        return ResultError(
            Result.Relay(
                d["fingerprint"],
                d["nickname"],
                d["address"],
                d["master_key_ed25519"],
                relay_in_recent_consensus=d.get(  # noqa
                    "relay_in_recent_consensus", None
                ),  # noqa
                relay_recent_measurement_attempt=d.get(  # noqa
                    "relay_recent_measurement_attempt", None
                ),  # noqa
                relay_recent_priority_list=d.get(  # noqa
                    "relay_recent_priority_list", None
                ),  # noqa
            ),
            d["circ"],
            d["dest_url"],
            d["scanner"],
            msg=d["msg"],
            t=d["time"],
        )

    def to_dict(self):
        d = super().to_dict()
        d.update({"msg": self.msg})
        return d


class ResultErrorCircuit(ResultError):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

    @property
    def type(self):
        return _ResultType.ErrorCircuit

    @property
    def freshness_reduction_factor(self):
        """
        There are a few instances when it isn't the relay's fault that the
        circuit failed to get built. Maybe someday we'll try detecting whose
        fault it most likely was and subclassing ResultErrorCircuit. But for
        now we don't. So reduce the freshness slightly more than ResultError
        does by default so priority isn't hurt quite as much.

        A (hopefully very very rare) example of when a circuit would fail to
        get built is when the sbws client machine suddenly loses Internet
        access.
        """
        return 0.6

    @staticmethod
    def from_dict(d):
        assert isinstance(d, dict)
        return ResultErrorCircuit(
            Result.Relay(
                d["fingerprint"],
                d["nickname"],
                d["address"],
                d["master_key_ed25519"],
                relay_in_recent_consensus=d.get(  # noqa
                    "relay_in_recent_consensus", None
                ),  # noqa
                relay_recent_measurement_attempt=d.get(  # noqa
                    "relay_recent_measurement_attempt", None
                ),  # noqa
                relay_recent_priority_list=d.get(  # noqa
                    "relay_recent_priority_list", None
                ),  # noqa
            ),
            d["circ"],
            d["dest_url"],
            d["scanner"],
            msg=d["msg"],
            t=d["time"],
        )

    def to_dict(self):
        d = super().to_dict()
        return d


class ResultErrorStream(ResultError):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

    @property
    def type(self):
        return _ResultType.ErrorStream

    @staticmethod
    def from_dict(d):
        assert isinstance(d, dict)
        return ResultErrorStream(
            Result.Relay(
                d["fingerprint"],
                d["nickname"],
                d["address"],
                d["master_key_ed25519"],
                relay_in_recent_consensus=d.get(  # noqa
                    "relay_in_recent_consensus", None
                ),  # noqa
                relay_recent_measurement_attempt=d.get(  # noqa
                    "relay_recent_measurement_attempt", None
                ),  # noqa
                relay_recent_priority_list=d.get(  # noqa
                    "relay_recent_priority_list", None
                ),  # noqa
            ),
            d["circ"],
            d["dest_url"],
            d["scanner"],
            msg=d["msg"],
            t=d["time"],
        )

    def to_dict(self):
        d = super().to_dict()
        return d


class ResultErrorSecondRelay(ResultError):
    """
    Error when it could not be found a second relay suitable to measure
    a relay.

    A second suitable relay is a relay that:
    - Has at least equal bandwidth as the relay to measure.
    - If the relay to measure is not an exit, \
      the second relay is an exit without `bad` flag and can exit to port 443.
    - If the relay to measure is an exit, the second relay is not an exit.

    It is instantiated in :func:`~sbws.core.scanner.measure_relay`.

    .. note:: this duplicates code and add more tech-debt,
       since it's the same as the other
       :class:`~sbws.lib.resultdump.ResultError` classes except for the
       ``type``.
       In a future refactor, there should be only one ``ResultError`` class
       and assign the type in the ``scanner`` module.
    """

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

    @property
    def type(self):
        return _ResultType.ErrorSecondRelay

    @staticmethod
    def from_dict(d):
        assert isinstance(d, dict)
        return ResultErrorSecondRelay(
            Result.Relay(
                d["fingerprint"],
                d["nickname"],
                d["address"],
                d["master_key_ed25519"],
                relay_in_recent_consensus=d.get(  # noqa
                    "relay_in_recent_consensus", None
                ),  # noqa
                relay_recent_measurement_attempt=d.get(  # noqa
                    "relay_recent_measurement_attempt", None
                ),  # noqa
                relay_recent_priority_list=d.get(  # noqa
                    "relay_recent_priority_list", None
                ),  # noqa
            ),
            d["circ"],
            d["dest_url"],
            d["scanner"],
            msg=d["msg"],
            t=d["time"],
        )

    def to_dict(self):
        d = super().to_dict()
        return d


class ResultErrorDestination(ResultError):
    """
    Error when there is not a working destination Web Server.

    It is instantiated in :func:`~sbws.core.scanner.measure_relay`.

    .. note:: this duplicates code and add more tech-debt,
       since it's the same as the other
       :class:`~sbws.lib.resultdump.ResultError` classes except for the
       ``type``.
       In a future refactor, there should be only one ``ResultError`` class
       and assign the type in the ``scanner`` module.
    """

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

    @property
    def type(self):
        return _ResultType.ErrorSecondRelay

    @staticmethod
    def from_dict(d):
        assert isinstance(d, dict)
        return ResultErrorSecondRelay(
            Result.Relay(
                d["fingerprint"],
                d["nickname"],
                d["address"],
                d["master_key_ed25519"],
                d["circ"],
                d["dest_url"],
                d["scanner"],
                relay_in_recent_consensus=d.get(  # noqa
                    "relay_in_recent_consensus", None
                ),  # noqa
                relay_recent_measurement_attempt=d.get(  # noqa
                    "relay_recent_measurement_attempt", None
                ),  # noqa
                relay_recent_priority_list=d.get(  # noqa
                    "relay_recent_priority_list", None
                ),  # noqa
            ),
            msg=d["msg"],
            t=d["time"],
        )

    def to_dict(self):
        d = super().to_dict()
        return d


class ResultErrorAuth(ResultError):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

    @property
    def type(self):
        return _ResultType.ErrorAuth

    @property
    def freshness_reduction_factor(self):
        """
        Override the default ResultError.freshness_reduction_factor because a
        ResultErrorAuth is most likely not the measured relay's fault, so we
        shouldn't hurt its priority as much. A higher reduction factor means a
        Result's effective freshness is reduced more, which makes the relay's
        priority better.

        The value 0.9 was chosen somewhat arbitrarily.
        """
        return 0.9

    @staticmethod
    def from_dict(d):
        assert isinstance(d, dict)
        return ResultErrorAuth(
            Result.Relay(
                d["fingerprint"],
                d["nickname"],
                d["address"],
                d["master_key_ed25519"],
                relay_in_recent_consensus=d.get(  # noqa
                    "relay_in_recent_consensus", None
                ),  # noqa
                relay_recent_measurement_attempt=d.get(  # noqa
                    "relay_recent_measurement_attempt", None
                ),  # noqa
                relay_recent_priority_list=d.get(  # noqa
                    "relay_recent_priority_list", None
                ),  # noqa
            ),
            d["circ"],
            d["dest_url"],
            d["scanner"],
            msg=d["msg"],
            t=d["time"],
        )

    def to_dict(self):
        d = super().to_dict()
        return d


class ResultSuccess(Result):
    def __init__(self, rtts, downloads, *a, **kw):
        super().__init__(*a, **kw)
        self._rtts = rtts
        self._downloads = downloads

    @property
    def type(self):
        return _ResultType.Success

    @property
    def rtts(self):
        return self._rtts

    @property
    def downloads(self):
        return self._downloads

    @staticmethod
    def from_dict(d):
        assert isinstance(d, dict)
        return ResultSuccess(
            d["rtts"] or [],
            d["downloads"],
            Result.Relay(
                d["fingerprint"],
                d["nickname"],
                d["address"],
                d["master_key_ed25519"],
                d["relay_average_bandwidth"],
                d.get("relay_burst_bandwidth"),
                d["relay_observed_bandwidth"],
                d.get("consensus_bandwidth"),
                d.get("consensus_bandwidth_is_unmeasured"),
                relay_in_recent_consensus=d.get(  # noqa
                    "relay_in_recent_consensus", None
                ),  # noqa
                relay_recent_measurement_attempt=d.get(  # noqa
                    "relay_recent_measurement_attempt", None
                ),  # noqa
                relay_recent_priority_list=d.get(  # noqa
                    "relay_recent_priority_list", None
                ),  # noqa
            ),
            d["circ"],
            d["dest_url"],
            d["scanner"],
            t=d["time"],
        )

    def to_dict(self):
        d = super().to_dict()
        d.update(
            {
                "rtts": self.rtts,
                "downloads": self.downloads,
                "relay_average_bandwidth": self.relay_average_bandwidth,
                "relay_burst_bandwidth": self.relay_burst_bandwidth,
                "relay_observed_bandwidth": self.relay_observed_bandwidth,
                "consensus_bandwidth": self.consensus_bandwidth,
                "consensus_bandwidth_is_unmeasured": self.consensus_bandwidth_is_unmeasured,  # noqa
            }
        )
        return d


class ResultDump:
    """Runs the enter() method in a new thread and collects new Results on its
    queue. Writes them to daily result files in the data directory"""

    def __init__(self, args, conf):
        assert os.path.isdir(conf.getpath("paths", "datadir"))
        self.conf = conf
        self.fresh_days = conf.getint("general", "data_period")
        self.datadir = conf.getpath("paths", "datadir")
        self.data = {}
        self.data_lock = RLock()
        self.thread = Thread(target=self.enter)
        self.queue = Queue()
        try:
            self.thread.start()
        except RuntimeError as e:
            fail_hard(e)

    def store_result(self, result):
        """Call from ResultDump thread"""
        assert isinstance(result, Result)
        with self.data_lock:
            fp = result.fingerprint
            if fp not in self.data:
                self.data[fp] = []
            self.data[fp].append(result)
            self.data = trim_results(self.fresh_days, self.data)
            # Not calling trim_results_ip_changed here to do not remove
            # the results for a relay that has changed address.
            # It will be called when loading the results to generate a v3bw
            # file.

    def handle_result(self, result):
        """Call from ResultDump thread. If we are shutting down, ignores
        ResultError* types"""
        assert isinstance(result, Result)
        fp = result.fingerprint
        nick = result.nickname
        if isinstance(result, ResultError) and settings.end_event.is_set():
            log.debug(
                "Ignoring %s for %s %s because we are shutting down",
                type(result).__name__,
                nick,
                fp,
            )
            return
        self.store_result(result)
        write_result_to_datadir(result, self.datadir)
        if result.type == "success":
            msg = (
                "Success measuring {} ({}) via circuit {} and "
                "destination {}".format(
                    result.fingerprint,
                    result.nickname,
                    result.circ,
                    result.dest_url,
                )
            )
        else:
            msg = (
                "Error measuring {} ({}) via circuit {} and "
                "destination {}: {}".format(
                    result.fingerprint,
                    result.nickname,
                    result.circ,
                    result.dest_url,
                    result.msg,
                )
            )
        # The result doesn't store the exit policies, so it can't be logged
        # whether it was an exit.
        if result.circ:
            as_exit = result.circ[1] == result.fingerprint
            msg += ". As exit." if as_exit else ". As entry."
        # When the error is that there are not more functional destinations.
        if result.type == "error-destination":
            log.warning(
                "Shutting down because there are not functional "
                "destinations."
            )
            # NOTE: Because this is executed in a thread, stop_threads can not
            # be call from here, it has to be call from the main thread.
            # Instead set the singleton end event, that will call stop_threads
            # from the main process.
            settings.end_event.set()
        # log every success/error measurement at debug level, to don't log the
        # the Web server at info level and because there's already a
        # heartbeat msg to indicate progress.
        log.debug(msg)

    def enter(self):
        """Main loop for the ResultDump thread.

        When there are results in the queue, queue.get will get them until
        there are not anymore or timeout happen.

        For every result it gets, it process it and store in the filesystem,
        which takes ~1 millisecond and will not trigger the timeout.
        It can then store in the filesystem ~1000 results per second.

        I does not accept any other data type than Results or list of Results,
        therefore is not possible to put big data types in the queue.

        If there are not any results in the queue, it waits 1 second and checks
        again.

        """
        with self.data_lock:
            self.data = load_recent_results_in_datadir(
                self.fresh_days, self.datadir
            )
        while not (settings.end_event.is_set() and self.queue.empty()):
            try:
                event = self.queue.get(timeout=1)
            except Empty:
                continue
            data = event
            if data is None:
                log.debug("Got None in ResultDump")
                continue
            elif isinstance(data, list):
                for r in data:
                    assert isinstance(r, Result)
                    self.handle_result(r)
            elif isinstance(data, Result):
                self.handle_result(data)
            else:
                log.warning(
                    "The only thing we should ever receive in the "
                    "result thread is a Result or list of Results. "
                    "Ignoring %s",
                    type(data),
                )

    def results_for_relay(self, relay):
        assert isinstance(relay, Relay)
        fp = relay.fingerprint
        with self.data_lock:
            if fp not in self.data:
                return []
            return self.data[fp]



#!/usr/bin/env python3
# Copyright 2019 juga (juga at riseup dot net), CC0 license.
"""Script to help release new versions.

usage: release.py GPG_KEY_IDENTIFIER

This is a fast and dirty script to help release new sbws versions.
It does not intent to cover all possible cases and errors because this is out
of the scope of sbws.
There are other tools [*] that help manage release tags, branches and
changelogs, but we should not add more requirements to sbws to just release a
new version.

This script will:
0. Detect the current program version
1. Ask which version to release
2. Update the changelog automatically with ``gitchangelog``
   and leave the option to user to manually edit it
3. Commit the changelog
4. Create a version with the tag and sign it
5. Push the commit and tag to the repository
6. Obtain the release tarball
7. Sign the release tarball
8. Modify the program to the next prerelease version and commit it
9. Push the commit

All is done in sequence and doesn't recover from any previous step.

It assumes that:
- the program version can be obtained with ``sbws.__version__``
- ``gitchangelog`` and ``semantic_version`` python packages are installed
- the official tarball releases are at gitlab.torproject.org
- the key to sign the release is only one and is available in the system

[*] See for instance https://semantic-release.gitbook.io,
    https://github.com/GitTools/GitVersion

"""
import subprocess
import sys
try:
    import semantic_version
except ImportError:
    print(
        "Please, install semantic_version: "
        "`apt install python3-semantic-version` or "
        "`pip install semantic_version`")
    sys.exit(1)
try:
    import gitchangelog  # noqa
except ImportError:
    print("Please, install gitchangelog: `pip install gitchangelog`")
    sys.exit(1)
# Not used directly, but gitchangelog depends on it
try:
    import mako  # noqa
except ImportError:
    print("Please, install mako: `pip install mako`")
    sys.exit(1)

import sbws


def obtain_release_version(version):
    release_type_dict = {
        1: version.next_patch(),
        2: version.next_minor(),
        3: version.next_major(),
    }
    print("Current prerelease version: ", version)
    print("which version would you like to release?:")
    print("1. to patch release: ", release_type_dict[1])
    print("2. to minor release: ", release_type_dict[2])
    print("3. to major release: ", release_type_dict[3])
    release_type = int(input())
    try:
        return release_type_dict[release_type]
    except KeyError:
        print("Invalid release.")
        sys.exit(1)


def main(args):
    print(__doc__)
    try:
        keyid = args[0]
    except IndexError:
        print("Please, pass a GPG key identifier as argument.")
        sys.exit(1)
    print("1. Which version to release")
    print("---------------------------")
    current_version = semantic_version.Version(sbws.__version__)
    release_version = obtain_release_version(current_version)

    print("\n2. Update changelog")
    print("-------------------")
    print("Creating tag v{} so that gitchangelog will create the new section "
          "with the correct new tag...".format(release_version))
    subprocess.call(['git', 'tag', 'v{}'.format(release_version),
                     '-m', '\"Release version {}.\"'.format(release_version)])

    print("\nUpdating the changelog automatically...")
    subprocess.call('gitchangelog'.split(' '))

    print("\nEdit the changelog manually to remove merge commits and "
          "leave only the apropiate paragraph for every bug/feature.")
    input("Press enter when done.\n")

    print("\nRemoving the tag...")
    subprocess.call(['git', 'tag', '-d', 'v{}'.format(release_version)])

    print("\n3. Commit the changelog")
    print("--------------------------")
    print("\nCommiting CHANGELOG.rst...")
    subprocess.call(['git', 'commit',
                     '-am', '"Release version {}."'.format(release_version)])

    print("\n4. Create tag and sign it")
    print("--------------------------")
    print("\nCreating the final tag and signing it...")
    subprocess.call(['git', 'tag', '-s', 'v{}'.format(release_version),
                     '-m', '"Release version {}."'.format(release_version)])

    print("\n5. Push commit and tag")
    print("------------------------")
    print("\nPush now so that the Gitlab creates the tarball from the new "
          " commit and tag, eg:")
    print("git push myremote mybranch")
    print("git push myremote --tags")
    input("Press enter when you are done.")

    print("\n6. Obtain the release tarball")
    print("-------------------------------")
    print("Obtaining Gitlab tarball...")
    subprocess.call(
        "wget https://gitlab.torproject.org/tpo/network-health/sbws/-/archive/v{}/sbws-v{}.tar.gz "  # noqa
        "-O v{}.tar.gz"
        .format(release_version, release_version, release_version).split(' ')
        )

    print("\n7. Create the tarball signature")
    print("-------------------------------")
    print("Creating detached signature...")
    subprocess.call("gpg --default-key {} "
                    "--output v{}.tar.gz.asc "
                    "--detach-sign v{}.tar.gz"
                    .format(keyid, release_version, release_version)
                    .split(' '))

    print("\nUpload the signature manually to Gitlab.")
    input("Press enter when done.")
    print("\nRelease done!!!.")

    print("\n8. Create next prerelease branch")
    print("----------------------------------")
    print("\nIf this release happens in a maintainance branch, merge the "
          "the commit to master and push, eg:"
          "git checkout master"
          "git merge --no-ff mybranch"
          "git push myremote master")
    next_branch_version = "maint{}".format(release_version)
    print("And create the next prerelease branch, eg:"
          "git checkout -b {}".format(next_branch_version))


if __name__ == "__main__":
    main(sys.argv[1:])



"""Unit tests for scanner.py."""
import concurrent.futures
import logging

import pytest
from freezegun import freeze_time

from sbws.core import scanner
from sbws.lib.circuitbuilder import CircuitBuilder
from sbws.lib.destination import DestinationList
from sbws.lib.heartbeat import Heartbeat
from sbws.lib.relayprioritizer import RelayPrioritizer

log = logging.getLogger(__name__)


def test_measurement_writer(
    sbwshome_only_datadir, result_success, rd, end_event
):
    if rd is None:
        pytest.skip("ResultDump is None")
    # Put one item in the queue
    scanner.measurement_writer(rd, result_success)
    assert rd.queue.qsize() == 1

    # Make queue maxsize 1, so that it'll be full after the first callback.
    # The second callback will wait 1 second, then the queue will be empty
    # again.
    rd.queue.maxsize = 1
    scanner.measurement_writer(rd, result_success)
    # after putting 1 result, the queue will be full
    assert rd.queue.qsize() == 1
    assert rd.queue.full()
    # it's still possible to put another results, because the callback will
    # wait 1 second and the queue will be empty again.
    scanner.measurement_writer(rd, result_success)
    assert rd.queue.qsize() == 1
    assert rd.queue.full()
    end_event.set()


def test_complete_measurements(
    args,
    conf,
    sbwshome_only_datadir,
    controller,
    relay_list,
    result_dump,
    rd,
    mocker,
):
    """
    Test that the ``ThreadPoolExecutor``` creates the epexted number of
    futures, ``process_completed_futures``process all of them and
    ``wait_futures_completed``  completes them if they were not already
    completed by the time ``process_completed_futures`` has already processed
    them.
    There are not real measurements done and the ``results`` are None objects.
    Running the scanner with the test network, test the real measurements.

    """
    with freeze_time("2020-02-29 10:00:00"):
        hbeat = Heartbeat(conf.getpath("paths", "state_fname"))
        # rl = RelayList(args, conf, controller, measurements_period, state)
        circuit_builder = CircuitBuilder(args, conf, controller, relay_list)
        # rd = ResultDump(args, conf)
        relay_prioritizer = RelayPrioritizer(args, conf, relay_list, rd)
        destinations, error_msg = DestinationList.from_config(
            conf, circuit_builder, relay_list, controller
        )
        num_threads = conf.getint("scanner", "measurement_threads")

        mocker.patch(
            "sbws.lib.destination.DestinationList.functional_destinations",
            side_effect=[d for d in destinations._all_dests],
        )
        print("start threads")
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=num_threads, thread_name_prefix="measurer"
        ) as executor:
            pending_results = {
                executor.submit(
                    scanner.dispatch_worker_thread,
                    args,
                    conf,
                    destinations,
                    circuit_builder,
                    relay_list,
                    target,
                ): target
                for target in relay_prioritizer.best_priority()
            }

            assert len(pending_results) == 321
            assert len(hbeat.measured_fp_set) == 0
            log.debug("Before process_completed_futures.")
            scanner.process_completed_futures(
                executor, hbeat, rd, pending_results
            )
            log.debug("After process_completed_futures")
            for pending_result in pending_results:
                assert pending_result.done() is True
            assert len(hbeat.measured_fp_set) == 321
            scanner.wait_futures_completed(pending_results)
            log.debug("After wait_futures_completed.")
            assert concurrent.futures.ALL_COMPLETED



# -*- coding: utf-8 -*-
"""Unit tests for resultdump."""

import datetime
import logging

from sbws.lib.relaylist import Relay
from sbws.lib.resultdump import (
    ResultError,
    ResultErrorStream,
    ResultSuccess,
    load_result_file,
    trim_results_ip_changed,
)


def test_trim_results_ip_changed_defaults(resultdict_ip_not_changed):
    results_dict = trim_results_ip_changed(resultdict_ip_not_changed)
    assert resultdict_ip_not_changed == results_dict


def test_trim_results_ip_changed_on_changed_ipv4_changed(
    resultdict_ip_changed, resultdict_ip_changed_trimmed
):
    results_dict = trim_results_ip_changed(
        resultdict_ip_changed, on_changed_ipv4=True
    )
    assert resultdict_ip_changed_trimmed == results_dict


def test_trim_results_ip_changed_on_changed_ipv4_no_changed(
    resultdict_ip_not_changed,
):
    results_dict = trim_results_ip_changed(
        resultdict_ip_not_changed, on_changed_ipv4=True
    )
    assert resultdict_ip_not_changed == results_dict


def test_trim_results_ip_changed_on_changed_ipv6(
    caplog, resultdict_ip_not_changed
):
    results_dict = trim_results_ip_changed(
        resultdict_ip_not_changed, on_changed_ipv6=True
    )
    assert resultdict_ip_not_changed == results_dict
    # There might be other logs from other threads.
    with caplog.at_level(logging.WARNING):
        assert (
            "Resetting bandwidth results when IPv6 changes, "
            "is not yet implemented.\n" in caplog.text
        )


def test_resultdump(
    rd, args, conf_results, controller, router_status, server_descriptor
):
    from sbws import settings

    relay = Relay(
        router_status.fingerprint,
        controller,
        ns=router_status,
        desc=server_descriptor,
    )
    relay.increment_relay_recent_priority_list()
    relay.increment_relay_recent_measurement_attempt()
    r = ResultSuccess(
        [],
        2000,
        relay,
        ["A", "B"],
        "http://localhost/bw",
        "scanner_nick",
    )
    # Storing the result with `rd.queue.put` will not store the result to disk
    # because the thread is not spawned with pytest.
    rd.store_result(r)
    results = rd.results_for_relay(relay)
    # It has stored the result
    assert 1 == len(results)
    # The result has the correct attribute
    assert 1 == len(results[0].relay_recent_priority_list)
    # Store a second result for the sme relay
    r = ResultError(
        relay,
        ["A", "B"],
        "http://localhost/bw",
        "scanner_nick",
    )
    rd.store_result(r)
    assert 2 == len(results)
    assert 1 == len(results[1].relay_recent_priority_list)
    settings.set_end_event()


def test_load(datadir):
    results = load_result_file(str(datadir.join("results.txt")))
    results = [v for values in results.values() for v in values]
    r1 = results[1]
    assert isinstance(r1, ResultSuccess)
    assert isinstance(
        r1.relay_recent_measurement_attempt[0], datetime.datetime
    )
    assert 2 == len(r1.relay_recent_measurement_attempt)
    assert 3 == len(r1.relay_recent_priority_list)
    assert 3 == len(r1.relay_in_recent_consensus)
    r2 = results[2]
    assert isinstance(r2, ResultErrorStream)
    assert isinstance(
        r2.relay_recent_measurement_attempt[0], datetime.datetime
    )
    assert 2 == len(r2.relay_recent_measurement_attempt)
    assert 3 == len(r2.relay_recent_priority_list)
    assert 3 == len(r2.relay_in_recent_consensus)
